-- See the Copyright notice at the end of this file.
--
expanded class DOUBLE
	--
	-- DOUBLE is the type for real numbers with more precision than 'REAL' type.
	-- The precision is not infinite.
	--
	-- Note: An Eiffel DOUBLE is mapped as a C double or as a Java double.
	--

inherit
	REAL_GENERAL
		export {} copy
		redefine fill_tagged_out_memory, out_in_tagged_out_memory, item, infix "<=", infix ">", infix ">="
		end

feature {ANY}
	infix "+" (other: like Current): DOUBLE is
		external "built_in"
		end

	infix "-" (other: like Current): DOUBLE is
		external "built_in"
		end

	infix "*" (other: like Current): DOUBLE is
		external "built_in"
		end

	infix "/" (other: like Current): DOUBLE is
		external "built_in"
		end

	infix "^" (e: INTEGER): DOUBLE is
			-- Raise Current to `e'-th power (see also `pow').
		external "built_in"
		end

	prefix "+": like Current is
		do
			Result := Current
		end

	prefix "-": DOUBLE is
		external "built_in"
		end

	abs: like Current is
		do
			if item < 0.0 then
				Result := - item
			else
				Result := item
			end
		end

	infix "<" (other: like Current): BOOLEAN is
		external "built_in"
		end

	infix "<=" (other: like Current): BOOLEAN is
		external "built_in"
		end

	infix ">" (other: like Current): BOOLEAN is
		external "built_in"
		end

	infix ">=" (other: like Current): BOOLEAN is
		external "built_in"
		end

	double_floor: DOUBLE is
		external "built_in"
		end

	floor: INTEGER is
			-- Greatest integral value no greater than Current.
		do
			Result := truncated_to_integer
		ensure
			result_no_greater: Result <= item
			close_enough: item - Result < one
		end

	double_ceiling: DOUBLE is
		do
			Result := item.ceil
		end

	ceiling: INTEGER is
			-- Smallest integral value no smaller than Current.
		local
			d: DOUBLE
		do
			d := double_ceiling
			Result := d.truncated_to_integer
		ensure
			result_no_smaller: Result >= item
			close_enough: Result - item < one
		end

	rounded: INTEGER is
		do
			if item.double_floor + 0.5 < item then
				Result := item.double_ceiling.truncated_to_integer
			else
				Result := item.double_floor.truncated_to_integer
			end
		end

	truncated_to_integer: INTEGER is
			-- Integer part (same sign, largest absolute value
			-- no greater than Current).
		external "built_in"
		ensure
			Result.to_double <= item
		end

	to_real: REAL is
			-- Note: C conversion from "double" to "float".
			-- Thus, Result can be undefine (ANSI C).
		external "built_in"
		end

	to_string: STRING is
			-- Convert the DOUBLE into a new allocated STRING.
			-- As ANSI C, the default number of digit for the
			-- fractionnal part is 6.
			-- Note: see `append_in' to save memory.
		do
			create Result.make(16)
			append_in(Result)
		end

	append_in (str: STRING) is
			-- Append the equivalent of `to_string' at the end of
			-- `str'. Thus you can save memory because no other
			-- STRING is allocate for the job.
		require
			str /= Void
		do
			append_in_format(str, 6)
		end

	to_string_format (f: INTEGER): STRING is
			-- Convert the DOUBLE into a new allocated STRING including
			-- only `f' digits in fractionnal part.
			-- Note: see `append_in_format' to save memory.
		require
			f >= 0
		do
			sprintf_double(sprintf_buffer, f, item)
			create Result.from_external_copy(sprintf_buffer.to_pointer)
		end

	append_in_format (str: STRING; f: INTEGER) is
			-- Same as `append_in' but produce `f' digit of
			-- the fractionnal part.
		require
			str /= Void
			f >= 0
		local
			i: INTEGER
		do
			from
				sprintf_double(sprintf_buffer, f, item)
				i := 0
			until
				sprintf_buffer.item(i) = '%U'
			loop
				str.extend(sprintf_buffer.item(i))
				i := i + 1
			end
		end

feature {ANY} -- Maths functions:
	sqrt: DOUBLE is
			-- Square root of `Current' (ANSI C `sqrt').
		require
			item >= 0
		external "built_in"
		end

	sin: DOUBLE is
			-- Sine of `Current' (ANSI C `sin').
		external "built_in"
		end

	cos: DOUBLE is
			-- Cosine of `Current' (ANSI C `cos').
		external "built_in"
		end

	tan: DOUBLE is
			-- Tangent of `Current' (ANSI C `tan').
		external "built_in"
		end

	asin: DOUBLE is
			-- Arc Sine of `Current' (ANSI C `asin').
		external "built_in"
		end

	acos: DOUBLE is
			-- Arc Cosine of `Current' (ANSI C `acos').
		external "built_in"
		end

	atan: DOUBLE is
			-- Arc Tangent of `Current' (ANSI C `atan').
		external "built_in"
		end

	atan2 (x: DOUBLE): DOUBLE is
			-- Arc Tangent of `Current' / `x'  (ANSI C `atan2').
		external "built_in"
		end

	sinh: DOUBLE is
			-- Hyperbolic Sine of `Current' (ANSI C `sinh').
		external "built_in"
		end

	cosh: DOUBLE is
			-- Hyperbolic Cosine of `Current' (ANSI C `cosh').
		external "built_in"
		end

	tanh: DOUBLE is
			-- Hyperbolic Tangent of `Current' (ANSI C `tanh').
		external "built_in"
		end

	exp: DOUBLE is
			-- Exponential of `Current' (ANSI C `exp').
		external "built_in"
		end

	log: DOUBLE is
			-- Natural Logarithm of `Current' (ANSI C `log').
		external "built_in"
		end

	log10: DOUBLE is
			-- Base-10 Logarithm of Current (ANSI C `log10').
		external "built_in"
		end

	pow (e: DOUBLE): DOUBLE is
			-- `Current' raised to the power of `e' (ANSI C `pow').
		external "built_in"
		end

feature {ANY} -- Object Printing:
	out_in_tagged_out_memory, fill_tagged_out_memory is
		do
			item.append_in(tagged_out_memory)
		end

feature {ANY} -- Miscellaneous:
	sign: INTEGER_8 is
			-- Sign of `Current' (0 -1 or 1).
		do
			if item < 0.0 then
				Result := -1
			elseif 0.0 < item then
				Result := 1
			else
				Result := 0
			end
		end

	one: DOUBLE is 1.0

	zero: DOUBLE is 0.0

	divisible (other: like Current): BOOLEAN is
		do
			Result := other.item /= 0.0
		end

feature {ANY} -- Hashing:
	hash_code: INTEGER is
		do
			Result := truncated_to_integer
			if Result < 0 then
				Result := - (Result + 1)
			end
		end

feature {ANY} -- For compatibility with the obsolete DOUBLE_REF style:
	item: DOUBLE

feature {REAL_GENERAL}
	ceil: DOUBLE is
		external "built_in"
		end

feature {}
	sprintf_double (buffer: NATIVE_ARRAY[CHARACTER]; f: INTEGER; d: DOUBLE) is
			-- Put in the `buffer' a viewable version of Current with
			-- `f' digit of the fractionnal part. Assume the `buffer' is
			-- large enougth.
		require
			f >= 0
		external "plug_in"
		alias "{
	 location: "${sys}runtime"
	 module_name: "basic_sprintf"
	 feature_name: "sprintf_double"
	 }"
		end

	sprintf_buffer: NATIVE_ARRAY[CHARACTER] is
		once
			Result := Result.calloc(1024)
		end

end -- class DOUBLE
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
