-- See the Copyright notice at the end of this file.
--
class AVL_DICTIONARY[V, K -> COMPARABLE]
	--
	-- Associative memory.
	-- Values of type `V' are stored using Keys of type `K'.
	--

inherit
	SIMPLE_DICTIONARY[V, K]
		redefine occurrences, fast_occurrences, key_at, fast_key_at
		end

insert
	AVL_HELPER[K]
		undefine
			copy, is_equal
		end

creation {ANY}
	make

feature {}
	root: AVL_DICTIONARY_NODE[V, K]

feature {ANY} -- Basic access:
	count: INTEGER is
		do
			if root /= Void then
				Result := root.count
			end
		end

	capacity: INTEGER is
		do
			Result := count --*** + lost_nodes.capacity
		end

	has (k: K): BOOLEAN is
			-- Is there a value currently associated with key `k'?
		do
			Result := count > 0 and then root.at(k) /= Void
		end

	fast_has (k: K): BOOLEAN is
			-- Is there a value currently associated with key `k'?
		do
			Result := count > 0 and then root.fast_at(k) /= Void
		end

	infix "@", at (k: K): V is
			-- Return the value associated to key `k'.
			-- (See also `reference_at' if V is a reference type.)
		do
			Result := root.at(k).value
		end

	fast_at (k: K): V is
		do
			Result := root.fast_at(k).value
		end

	reference_at (k: K): V is
			-- Return Void or the value associated with key `k'. Actually, this
			-- feature is useful when the type of values (the type V) is a
			-- reference type, to avoid using `has' just followed by `at' to get
			-- the corresponding value.
		local
			n: AVL_DICTIONARY_NODE[V, K]
		do
			if root /= Void then
				n := root.at(k)
				if n /= Void then
					Result := n.value
				end
			end
		end

	fast_reference_at (k: K): V is
			-- Return Void or the value associated with key `k'. Actually, this
			-- feature is useful when the type of values (the type V) is a
			-- reference type, to avoid using `has' just followed by `at' to get
			-- the corresponding value.
		local
			n: AVL_DICTIONARY_NODE[V, K]
		do
			if root /= Void then
				n := root.fast_at(k)
				if n /= Void then
					Result := n.value
				end
			end
		end

feature {ANY}
	put (v: V; k: K) is
			-- Change some existing entry or `add' the new one. If there is
			-- as yet no key `k' in the dictionary, enter it with item `v'.
			-- Otherwise overwrite the item associated with key `k'.
		local
			node, r: like root
		do
			if root = Void then
				root := new_node(k)
				root.set_value(v)
			else
				node := root.at(k)
				if node = Void then
					r := root.add(k)
					node := root.last_inserted
					root := r
				end
				node.set_value(v)
			end
		end

	add (v: V; k: K) is
			-- To add a new entry `k' with its associated value `v'. Actually,
			-- this is equivalent to call `put', but may run a little bit
			-- faster.
		local
			r: like root
		do
			if root = Void then
				root := new_node(k)
				root.set_value(v)
			else
				r := root.add(k)
				root.last_inserted.set_value(v)
				root := r
			end
		end

feature {ANY} -- Looking and searching some value:
	occurrences (v: V): INTEGER is
			-- Number of occurrences using `equal'.
			-- See also `fast_occurrences' to chose the apropriate one.
		do
			if root /= Void then
				Result := root.occurrences(v)
			end
		end

	fast_occurrences (v: V): INTEGER is
			-- Number of occurrences using `='.
			-- See also `occurrences' to chose the apropriate one.
		do
			if root /= Void then
				Result := root.fast_occurrences(v)
			end
		end

	key_at (v: V): K is
			-- Retrieve the key used for value `v' using `equal' for comparison.
		do
			Result := root.key_at(v)
		end

	fast_key_at (v: V): K is
			-- Retrieve the key used for value `v' using `=' for comparison.
		do
			Result := root.fast_key_at(v)
		end

feature {ANY} -- Removing:
	remove (k: K) is
			-- Remove entry `k' (which may exist or not before this call).
		local
			safe_equal: SAFE_EQUAL[K]
		do
			if root /= Void then
				if count = 1 and then safe_equal.test(root.key, k) then
					discard_node(root)
					root := Void
				else
					root := root.remove(k)
				end
			end
		end

	clear_count is
			-- Discard all items.
		do
			if not is_empty then
				root.clear_nodes
				root := Void
			end
		end

	clear_count_and_capacity is
		do
			clear_count
			--*** lost_nodes.clear_count_and_capacity
		end

feature {ANY} -- To provide iterating facilities:
	map: FAST_ARRAY[AVL_DICTIONARY_NODE[V, K]]
			-- Elements in a row for iteration. See `build_map'.

	map_dirty: BOOLEAN
			-- True when the map needs to be built again for the iterators. See
			-- `build_map'.

	item (index: INTEGER): V is
		do
			if map_dirty then
				build_map
			end
			Result := map.item(index - 1).value
		end

	key (index: INTEGER): K is
		do
			if map_dirty then
				build_map
			end
			Result := map.item(index - 1).key
		end

	get_new_iterator_on_keys: ITERATOR[K] is
		do
			create {ITERATOR_ON_DICTIONARY_KEYS[K]} Result.make(Current)
		end

feature {} -- Iterating internals:
	build_map is
		require
			build_needed: map_dirty
		do
			map.clear
			if count > 0 then
				root.map_in(map)
			end
			map_dirty := False
		end

feature {ANY} -- Other features:
	internal_key (k: K): K is
			-- Retrieve the internal key object which correspond to the existing
			-- entry `k' (the one memorized into the `Current' dictionary).
		do
			Result := root.at(k).key
		end

feature {}
	make is
		do
		end

feature {} -- from AVL_HELPER
	node_memo: MEMO[AVL_DICTIONARY_NODE[V, K]] is
      do
         -- Result is Void
		end

	inserted_node_memo: MEMO[AVL_DICTIONARY_NODE[V, K]] is
      do
         -- Result is Void
		end

	lost_nodes: COLLECTION[AVL_DICTIONARY_NODE[V, K]] is
      do
         -- Result is Void
		end

	a_new_node (k: K): AVL_DICTIONARY_NODE[V, K] is
		do
			create Result.make(k)
		end

end -- class AVL_DICTIONARY
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
