-- See the Copyright notice at the end of this file.
--
deferred class AVL_NODE[E -> COMPARABLE]
	--
	-- Auxiliary class to implement AVL_SET[E].
	--
	-- This a classic implementation of an AVL tree (balanced tree first
	-- designed by Adelson-Velskii and Landis, 1960)
	--

inherit
	COMPARABLE
		--insert
		--   AVL_HELPER[E]
		--      redefine
		--         out_in_tagged_out_memory
		--      end
		undefine is_equal
		redefine out_in_tagged_out_memory
		end

feature {ANY}
	infix "<" (other: like Current): BOOLEAN is
		do
			Result := item < other.item
		end

	out_in_tagged_out_memory is
		do
			item.out_in_tagged_out_memory
			tagged_out_memory.extend('(')
			if left = Void then
				tagged_out_memory.append(once "Void")
			else
				left.out_in_tagged_out_memory
			end
			tagged_out_memory.extend(',')
			if right = Void then
				tagged_out_memory.append(once "Void")
			else
				right.out_in_tagged_out_memory
			end
			tagged_out_memory.extend(')')
		end

feature {AVL_NODE, AVL_HELPER}
	item: E

	left: like Current

	right: like Current

	count: INTEGER is
		do
			Result := 1
			if left /= Void then
				Result := Result + left.count
			end
			if right /= Void then
				Result := Result + right.count
			end
		end

feature {AVL_NODE, AVL_HELPER}
	last_inserted: like Current is
		do
			Result := last_inserted_memory.item
		end

	add (i: like item): like Current is
			-- Returns the root node that should replace this one
		local
			dummy: BOOLEAN
		do
			make_new_node(i)
			dummy := do_insert
			Result := node
			if node_inserted.item then
				clear_node
			end
		ensure
			last_inserted.item = i
			Result /= Void
		end

	remove (i: like item): like Current is
			-- Returns the root node that should replace this one
		local
			dummy: BOOLEAN
		do
			make_new_node(i)
			dummy := do_remove
			Result := node
			clear_node
		ensure
			lost_nodes.count = old lost_nodes.count + 1
			Result /= Void
		end

	clear_nodes is
		do
			if left /= Void then
				left.clear_nodes
			end
			if right /= Void then
				right.clear_nodes
			end
			discard_node(Current)
		end

	set_item (i: like item) is
		require
			i /= Void
		do
			item := i
		ensure
			item = i
		end

	map_in (map: COLLECTION[like Current]) is
		require
			map /= Void
		do
			if left /= Void then
				left.map_in(map)
			end
			map.add_last(Current)
			if right /= Void then
				right.map_in(map)
			end
		ensure
			map.count = old map.count + count
		end

feature {AVL_NODE}
	balanced: INTEGER is 0

	imbalanced_left: INTEGER is -1

	imbalanced_right: INTEGER is 1

	balance: INTEGER
			-- Balance factor; either `balanced' (the tree is balanced),
			-- `imbalanced_left' (the left branch is the longer) or
			-- `imbalanced_right' (the right branch is the longer)

	set_balance (b: like balance) is
		do
			balance := b
		ensure
			balance = b
		end

	set_left (l: like left) is
		do
			left := l
		ensure
			left = l
		end

	set_right (r: like right) is
		do
			right := r
		ensure
			right = r
		end

feature {AVL_NODE} -- Insertion:
	do_insert: BOOLEAN is
			-- True if some reorganisation may be needed above the current node.
		local
			n: like node
		do
			n := node
			if n < Current then
				if insert_left then
					Result := left_grown
				end
			elseif Current < n then
				if insert_right then
					Result := right_grown
				end
			else
				Result := False -- already in the tree...
			end
		end

	left_grown: BOOLEAN is
		local
			n: like node
		do
			n := node
			inspect
				n.balance
			when imbalanced_right then
				n.set_balance(balanced)
			when balanced then
				n.set_balance(imbalanced_left)
				Result := True
			when imbalanced_left then
				if n.left.balance = imbalanced_left then
					n.set_balance(balanced)
					n.left.set_balance(balanced)
					set_node(n.rotate_right)
				else
					inspect
						n.left.right.balance
					when imbalanced_left then
						n.set_balance(imbalanced_right)
						n.left.set_balance(balanced)
					when balanced then
						n.set_balance(balanced)
						n.left.set_balance(balanced)
					when imbalanced_right then
						n.set_balance(balanced)
						n.left.set_balance(imbalanced_left)
					end
					n.left.right.set_balance(balanced)
					n.set_left(n.left.rotate_left)
					set_node(n.rotate_right)
				end
			end
		end

	right_grown: BOOLEAN is
		local
			n: like node
		do
			n := node
			inspect
				n.balance
			when imbalanced_left then
				n.set_balance(balanced)
			when balanced then
				n.set_balance(imbalanced_right)
				Result := True
			when imbalanced_right then
				if n.right.balance = imbalanced_right then
					n.set_balance(balanced)
					n.right.set_balance(balanced)
					set_node(n.rotate_left)
				else
					inspect
						n.right.left.balance
					when imbalanced_left then
						n.set_balance(balanced)
						n.right.set_balance(imbalanced_right)
					when balanced then
						n.set_balance(balanced)
						n.right.set_balance(balanced)
					when imbalanced_right then
						n.set_balance(imbalanced_left)
						n.right.set_balance(balanced)
					end
					n.right.left.set_balance(balanced)
					n.set_right(n.right.rotate_right)
					set_node(n.rotate_left)
				end
			end
		end

	insert_right: BOOLEAN is
		local
			n: like Current
		do
			if right = Void then
				n := node
				right := n
				n.set_balance(balanced)
				Result := True
				set_node_inserted
			else
				Result := right.do_insert
			end
			if Result then
				set_node(Current)
			end
		end

	insert_left: BOOLEAN is
		local
			n: like Current
		do
			if left = Void then
				n := node
				left := n
				n.set_balance(balanced)
				Result := True
				set_node_inserted
			else
				Result := left.do_insert
			end
			if Result then
				set_node(Current)
			end
		end

feature {AVL_NODE} -- Removal:
	do_remove: BOOLEAN is
		local
			n: like node; b: BOOLEAN; f: TUPLE[BOOLEAN, BOOLEAN]
		do
			n := node
			if n < Current then
				Result := n.left.do_remove and then n.left_shrunk
			elseif n > Current then
				Result := n.right.do_remove and then n.right_shrunk
			else
				f := [False, False]
				b := True
				if n.left /= Void then
					n.set_left(n.find_highest(n.left, f))
					if f.second then
						if f.first then
							Result := n.left_shrunk
						end
						b := False
					end
				end
				if b and then n.right /= Void then
					f.set_first(False)
					f.set_second(False)
					n.set_right(n.find_lowest(n.right, f))
					if f.second then
						if f.first then
							Result := n.right_shrunk
						end
						b := False
					end
				end
				if b then
					discard_node(Current)
					Result := True
				end
			end
		end

	left_shrunk: BOOLEAN is
		local
			n: like node
		do
			n := node
			inspect
				n.balance
			when imbalanced_left then
				n.set_balance(balanced)
				Result := True
			when balanced then
				n.set_balance(imbalanced_right)
			when imbalanced_right then
				inspect
					n.right.balance
				when imbalanced_left then
					inspect
						n.right.left.balance
					when imbalanced_left then
						n.set_balance(balanced)
						n.right.set_balance(imbalanced_right)
					when balanced then
						n.set_balance(balanced)
						n.right.set_balance(balanced)
					when imbalanced_right then
						n.set_balance(imbalanced_left)
						n.right.set_balance(balanced)
					end
					n.right.left.set_balance(balanced)
					n.set_right(n.right.rotate_right)
					set_node(n.rotate_left)
					Result := True
				when balanced then
					n.set_balance(imbalanced_right)
					n.right.set_balance(imbalanced_left)
					set_node(n.rotate_left)
				when imbalanced_right then
					n.set_balance(balanced)
					n.right.set_balance(balanced)
					set_node(n.rotate_left)
				end
			end
		end

	right_shrunk: BOOLEAN is
		local
			n: like node
		do
			n := node
			inspect
				n.balance
			when imbalanced_left then
				inspect
					n.left.balance
				when imbalanced_left then
					n.set_balance(balanced)
					n.left.set_balance(balanced)
					set_node(n.rotate_right)
					Result := True
				when balanced then
					n.set_balance(imbalanced_left)
					n.left.set_balance(imbalanced_right)
					set_node(n.rotate_right)
				when imbalanced_right then
					inspect
						n.left.right.balance
					when imbalanced_left then
						n.set_balance(imbalanced_right)
						n.left.set_balance(balanced)
					when balanced then
						n.set_balance(balanced)
						n.left.set_balance(balanced)
					when imbalanced_right then
						n.set_balance(balanced)
						n.left.set_balance(imbalanced_left)
					end
					n.left.right.set_balance(balanced)
					n.set_left(n.left.rotate_left)
					set_node(n.rotate_right)
					Result := True
				end
			when balanced then
				n.set_balance(imbalanced_left)
			when imbalanced_right then
				n.set_balance(balanced)
				Result := True
			end
		end

	find_highest (n: like Current; flags: TUPLE[BOOLEAN, BOOLEAN]): like Current is
			-- flags.first is True if the tree must be balanced
			-- flags.second is True if a node was found that is higher that Current
		do
			Result := n
			if Result = Void then
				flags.set_second(False)
			else
				if Result.right /= Void then
					Result := find_highest(Result.right, flags)
					if flags.second and then flags.first then
						flags.set_first(n.right_shrunk)
					end
				else
					set_item(Result.item)
					Result := Result.left
					discard_node(n)
				end
			end
		end

	find_lowest (n: like Current; flags: TUPLE[BOOLEAN, BOOLEAN]): like Current is
			-- flags.first is True if the tree must be balanced
			-- flags.second is True if a node was found that is higher that Current
		do
			Result := n
			if Result = Void then
				flags.set_second(False)
			else
				if Result.left /= Void then
					Result := find_lowest(Result.left, flags)
					if flags.second and then flags.first then
						flags.set_first(n.right_shrunk)
					end
				else
					set_item(Result.item)
					Result := Result.right
					lost_nodes.add_last(n)
				end
			end
		end

feature {AVL_NODE} -- Rotations:
	rotate_right: like Current is
			-- Proceeds to some reorganisation and returns the upper node.
		local
			left_right: like Current
		do
			Result := left
			left_right := left.right
			left.set_right(Current)
			set_left(left_right)
		end

	rotate_left: like Current is
			-- Proceeds to some reorganisation and returns the upper node.
		local
			right_left: like Current
		do
			Result := right
			right_left := right.left
			right.set_left(Current)
			set_right(right_left)
		end

feature {} -- Creation:
	make (i: like item) is
		require
			i /= Void
		do
			set_item(i)
			set_balance(balanced)
		ensure
			item = i
		end

feature {}
	node_inserted: reference BOOLEAN is
		once
			create Result
		end

	node: like Current is
			-- Used when inserting a new node while avoiding memory leaks.
		do
			Result := node_memo.item
		end

	make_new_node (i: like item) is
		local
			n: like node
		do
			n := inserted_node_memo.item
			if n = Void then
				n := new_node(i)
				inserted_node_memo.set_item(n)
			else
				n.set_item(i)
			end
			n.set_left(Void)
			n.set_right(Void)
			n.set_balance(balanced)
			node_memo.set_item(n)
			inserted(n)
		ensure
			node.item = i
		end

	set_node (n: like node) is
		do
			node_memo.set_item(n)
		ensure
			node = n
		end

	clear_node is
			-- Used when the node has truly been inserted.
		do
			node_memo.set_item(Void)
			inserted_node_memo.set_item(Void)
		ensure
			node = Void
		end

	set_node_inserted is
		do
			node_inserted.set_item(True)
		ensure
			node_inserted.item
		end

	clear_node_inserted is
		do
			node_inserted.set_item(False)
		ensure
			not node_inserted.item
		end

feature {}
	node_memo: MEMO[AVL_NODE[E]] is
			-- Holds `node'. The held node may change during insertion
      deferred
		end

	inserted_node_memo: MEMO[AVL_NODE[E]] is
			-- A node about to be inserted. If the node, is truly inserted, then
			-- that memory will be cleared; otherwise the node is reused at next
			-- insertion (see `make_new_node')
      deferred
		end

	lost_nodes: COLLECTION[AVL_NODE[E]] is
			-- Keeps discarded nodes to avoid memory leaks
		deferred
		end

	new_node (e: like item): like Current is
		deferred
		end

	discard_node (n: like Current) is
		deferred
		end

feature {}
	inserted (n: like Current) is
		do
			last_inserted_memory.set_item(n)
		end

	last_inserted_memory: MEMO[like Current] is
		deferred
		end

invariant
	left /= Void implies left < Current

	right /= Void implies Current < right

end -- class AVL_NODE
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
