-- See the Copyright notice at the end of this file.
--
class AVL_SET[E -> COMPARABLE]
	--
	-- Definition of a mathematical set of comparable objects. All common
	-- operations on mathematical sets are available.
	--

inherit
	SET[E]

insert
	AVL_HELPER[E]
		undefine
			copy, is_equal
		end
	SAFE_EQUAL[E]
		undefine
			copy, is_equal
		end

creation {ANY}
	make, from_collection

feature {AVL_SET}
	root: AVL_SET_NODE[E]

feature {}
	make is
		do
			create map.make(0)
			map_dirty := False
		end

feature {ANY}
	debug_string: STRING is
		do
			if root = Void then
				Result := once "Void"
			else
				Result := tagged_out_memory
				Result.clear_count
				root.out_in_tagged_out_memory
			end
		end

feature {ANY} -- Adding and removing:
	add (e: like item) is
			-- Add a new item to the set: the mathematical definition of
			-- adding in a set is followed.
		do
			if root = Void then
				check is_empty end
				root := new_node(e)
				count := 1
				map_dirty := True
			elseif not root.has(e) then
				if count = 0 then
					root := new_node(e)
				elseif not has(e) then
					root := root.add(e)
				end
				count := count + 1
				map_dirty := True
			end
		end

	remove (e: like item) is
			-- Remove item `e' from the set: the mathematical definition of
			-- removing from a set is followed.
		do
			if has(e) then
				root := root.remove(e)
				count := count - 1
				map_dirty := True
			end
		end

	clear_count, clear_count_and_capacity is
		do
			if not is_empty then
				root.clear_nodes
				root := Void
				count := 0
				map_dirty := True
			end
		end

feature {ANY} -- Looking and searching:
	has (e: like item): BOOLEAN is
			-- Is element `e' in the set?
		do
			Result := root /= Void and then root.has(e)
		end

	fast_has (e: like item): BOOLEAN is
			-- Is element `e' in the set?
		do
			Result := root /= Void and then root.fast_has(e)
		end

	reference_at (e: like item): like item is
		do
			if root /= Void then
				Result := root.at(e).item
			end
		end

feature {ANY} -- To provide iterating facilities:
	item (index: INTEGER): E is
			-- Return the item indexed by `index'.
		do
			if map_dirty then
				build_map
			end
			Result := map.item(index - 1).item
		end

feature {} -- Iterating internals:
	build_map is
		require
			build_needed: map_dirty
		do
			map.clear
			if count > 0 then
				root.map_in(map)
			end
			map_dirty := False
		ensure
			build_done: not map_dirty
		end

	map: FAST_ARRAY[AVL_SET_NODE[E]]
			-- Elements in a row for iteration. See `build_map'.

	map_dirty: BOOLEAN
			-- True when the map needs to be built again for the iterators. See
			-- `build_map'.

feature {} -- from AVL_HELPER
	node_memo: MEMO[AVL_SET_NODE[E]] is
      do
         -- Result is Void
		end

	inserted_node_memo: MEMO[AVL_SET_NODE[E]] is
      do
         -- Result is Void
		end

	lost_nodes: COLLECTION[AVL_SET_NODE[E]] is
      do
         -- Result is Void
		end

	a_new_node (e: E): AVL_SET_NODE[E] is
		do
			create Result.make(e)
		end

invariant
	not map_dirty implies map.count = count

	count > 0 implies root /= Void and then root.count = count

end -- class AVL_SET
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
