/*
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
*/
/*
  This file (SmartEiffel/sys/runtime/profile.c) is automatically included when
  profiling (option -profile)
*/

/*
  WARNING: the profile is currently incompatible with SCOOP
  (not thread-safe)
*/

#define NANO_SEC INT64_C(1000000000) /* 1 second is that many nanoseconds */
#define NANO_2_SEC INT64_C(2000000000) /* 2 seconds */

/*
 * We start by the definition of the current_time() function. It is the only
 * one that is OS-dependant. If your OS does not appear in the list, please be
 * kind and try to implement current_time() for your OS, and send a patch to
 * smarteiffel@loria.fr.
 *
 * Currently supported OSes:
 *
 *  - POSIX (UNIX, Linux and the like; also surely MacOS X and WinNT)
 *
 *  - Windows (any 32-bit flavour)
 */

static long current_time_counter = 0L;

/* -- [ POSIX ] --------------------------------------------------------- */
#if defined _POSIX_VERSION

#include <sys/time.h>
#include <time.h>

static int posix_current_time(se_time_t* result) {
  struct timeval res;
  int r = !gettimeofday(&res, NULL);
  if (r) {
    result->se_seconds = (int64_t)res.tv_sec;
    result->se_nanoseconds = (int64_t)res.tv_usec * INT64_C(1000);
    current_time_counter++;
  }
  if (!r) fprintf(SE_ERR, "*** WARNING: current_time() failed\n");
  return r;
}

/* *** to be fixed
#  if defined CLOCK_THREAD_CPUTIME_ID
#    define PROFILE_CLOCK CLOCK_THREAD_CPUTIME_ID
#  elif defined CLOCK_PROCESS_CPUTIME_ID
#    define PROFILE_CLOCK CLOCK_PROCESS_CPUTIME_ID
#  elif defined CLOCK_MONITONIC
#    define PROFILE_CALOCK CLOCK_MONOTONIC
#  elif defined CLOCK_REALTIME
#    define PROFILE_CLOCK CLOCK_REALTIME
#  endif
*/
#  if defined CLOCK_REALTIME
#    define PROFILE_CLOCK CLOCK_REALTIME
#  endif

#  ifdef PROFILE_CLOCK

static int current_time(se_time_t* result) {
  struct timespec res;
  int r = !clock_gettime(PROFILE_CLOCK, &res);
  if (r) {
    result->se_seconds = (int64_t)res.tv_sec;
    result->se_nanoseconds = (int64_t)res.tv_nsec;
    current_time_counter++;
  }
  else {
    r = posix_current_time(result);
  }

  return r;
}

#  else

/* POSIX, but no real-time timers: use gettimeofday() */

#    define current_time posix_current_time

#  endif

/* -- [ Windows ] ------------------------------------------------------- */
#elif defined WIN32

/*#include <windows.h> done in base.h */

static int current_time(se_time_t* result) {
  int64_t freq, ticks;
  int r = QueryPerformanceFrequency(&freq);
  if (r) {
    if (freq <= 0) {
      r = 0;
    } else {
      r = QueryPerformanceCounter(&ticks);
      if (r) {
	result->se_seconds = ticks / freq;
	result->se_nanoseconds = (int64_t)((NANO_SEC*ticks) % freq);
	current_time_counter++;
      }
    }
  }
  if (!r) fprintf(SE_ERR, "*** WARNING: current_time() failed\n");
  return r;
}

/* -- [ Others ] -------------------------------------------------------- */
#else

/* No known time function; times will stay null. */
static int current_time(se_time_t* result) {
  result->se_seconds = INT64_C(0);
  result->se_nanoseconds = INT64_C(0);
  current_time_counter++;
  return 1;
}

#endif

static int divide_time(se_time_t* time, int divisor) {
  int r = 0;
  int64_t s = time->se_seconds;
  int64_t n = time->se_nanoseconds;

  if (divisor != 0) {
    s /= divisor;
    n /= divisor;
    time->se_seconds = s;
    time->se_nanoseconds = n;
    r = 1;
  }

  if (!r) {
    fprintf(SE_ERR, "*** WARNING: could not divide time (%" PRId64 ".%.9" PRId64 ") by %i\n", s, n, divisor);
  }

  return r;
}

static int multiply_time(se_time_t* time, int mult) {
  int r = 0;
  int64_t s = time->se_seconds;
  int64_t n = time->se_nanoseconds;

  s *= mult;
  n *= mult;

  if (n > -(NANO_SEC*2) && n < (NANO_SEC*2)) {
    while (n < INT64_C(0) && s > INT64_C(0)) {
      --s;
      n += NANO_SEC;
    }
    while (n >= NANO_SEC) {
      ++s;
      n -= NANO_SEC;
    }
  }

  if (s >= INT64_C(0) && n >= INT64_C(0) && n < NANO_SEC) {
    time->se_seconds = s;
    time->se_nanoseconds = n;
    r = 1;
  }

  if (!r) {
    fprintf(SE_ERR, "*** WARNING: could not multiply time (%" PRId64 ".%.9" PRId64 ") by %i\n", s, n, mult);
  }

  return r;
}

static int subtract_time(se_time_t* start, se_time_t* stop, se_time_t* result) {
  int r = 0;
  int64_t s = stop->se_seconds - start->se_seconds;
  int64_t n = stop->se_nanoseconds - start->se_nanoseconds;

  if (n < INT64_C(0)) {
    --s;
    n += NANO_SEC;
  }
  else if (n >= NANO_SEC) {
    ++s;
    n -= NANO_SEC;
  }
  if (s >= INT64_C(0) && n >= INT64_C(0) && n < NANO_SEC) {
    result->se_seconds = s;
    result->se_nanoseconds = n;
    r = 1;
  }

  if (!r) {
    fprintf(SE_ERR, "*** WARNING: subtract_time failed (%" PRId64 ".%.9" PRId64 ")\n", s, n);
    fprintf(SE_ERR, "                            (start %" PRId64 ".%.9" PRId64 ")\n", start->se_seconds, start->se_nanoseconds);
    fprintf(SE_ERR, "                            (stop  %" PRId64 ".%.9" PRId64 ")\n", stop->se_seconds, stop->se_nanoseconds);
  }

  return r;
}

static int add_time(se_time_t* first, se_time_t* second, se_time_t* result) {
  int r = 0;
  int64_t s = first->se_seconds + second->se_seconds;
  int64_t n = first->se_nanoseconds + second->se_nanoseconds;

  if (n > -(NANO_SEC*2) && n < (NANO_SEC*2)) {
    while (n < INT64_C(0) && s > INT64_C(0)) {
      --s;
      n += NANO_SEC;
    }
    while (n >= NANO_SEC) {
      ++s;
      n -= NANO_SEC;
    }
  }

  if (s >= INT64_C(0) && n >= INT64_C(0) && n < NANO_SEC) {
    result->se_seconds = s;
    result->se_nanoseconds = n;
    r = 1;
  }

  if (!r) {
    fprintf(SE_ERR, "*** WARNING: add_time failed (%" PRId64 ".%.9" PRId64 ")\n", s, n);
  }

  return r;
}

/* ------------------------------------------------------------------------ */

static se_profile_access_t* _profile_access(se_profile_access_t* access, se_profile_t* profile) {
  se_profile_access_t* result = NULL;
  se_profile_access_t* point = access;

  if (point != NULL) {
    if (point->access == profile) {
      result = point;
    }
    else {
      do {
        point = point->next;
	if (point->access == profile) {
	  result = point;
	}
      } while (result == NULL && point != access);
    }
  }

  return result;
}

/* --*** The two following functions are quite similar. Some C wizard
 * is welcome here to further factorize the code.
 */

static se_profile_source_access_t* profile_source_access(se_profile_source_access_t** access_list, se_profile_t* profile) {
  se_profile_source_access_t* access = *access_list;
  se_profile_source_access_t* prev;
  se_profile_source_access_t* result = (se_profile_source_access_t*)_profile_access((se_profile_access_t*)access, profile);

  if (result != NULL) {
    prev = (se_profile_source_access_t*)result->prev;
    if ((result != access) && (result != prev) &&
	((result->count > (INT64_C(1) + prev->count)))) {
      /* kinda sorting using the number of calls (heuristics to find
	 most quickly the most used features */
      result->prev = prev->prev;
      prev->next = result->next;
      result->next = prev;
      prev->prev = result;
      result->prev->next = result;
      prev->next->prev = prev;
      if (prev == access) {
        *access_list = result;
      }
    }
  }
  else {
    result = (se_profile_source_access_t*)se_malloc(sizeof(se_profile_source_access_t));
    result->access = profile;
    result->count = INT64_C(0);

    if (access == NULL) {
      *access_list = result->next = result->prev = result;
    }
    else {
      prev = access->prev;
      result->next = access;
      access->prev = result;
      result->prev = prev;
      prev->next   = result;
    }
  }

  return result;
}

static se_profile_target_access_t* profile_target_access(se_profile_target_access_t** access_list, se_profile_t* profile) {
  se_profile_target_access_t* access = *access_list;
  se_profile_target_access_t* prev;
  se_profile_target_access_t* result = (se_profile_target_access_t*)_profile_access((se_profile_access_t*)access, profile);

  if (result != NULL) {
    prev = (se_profile_target_access_t*)result->prev;
    if ((result != access) && (result != prev) &&
	((result->count > (INT64_C(1) + prev->count)))) {
      /* kinda sorting using the number of calls (heuristics to find
	 most quickly the most used features */
      result->prev = prev->prev;
      prev->next = result->next;
      result->next = prev;
      prev->prev = result;
      result->prev->next = result;
      prev->next->prev = prev;
      if (prev == access) {
        *access_list = result;
      }
    }
  }
  else {
    result = (se_profile_target_access_t*)se_malloc(sizeof(se_profile_target_access_t));
    result->access = profile;
    result->count = INT64_C(0);
    result->time.se_seconds = INT64_C(0);
    result->time.se_nanoseconds = INT64_C(0);

    if (access == NULL) {
      *access_list = result->next = result->prev = result;
    }
    else {
      prev = access->prev;
      result->next = access;
      access->prev = result;
      result->prev = prev;
      prev->next   = result;
    }
  }

  return result;
}

#define time_comparison(left, right) (  ((left)->se_seconds==(right)->se_seconds)          \
                                      ? ((left)->se_nanoseconds - (right)->se_nanoseconds) \
                                      : ((left)->se_seconds - (right)->se_seconds))

static int profile_nest = 0;
static se_profile_t profiler_profile;
static se_time_t measure_drift;

/* ====================================================================== */

/*
 * Below are some functions used by the Eiffel runtime to profile the system.
 */

void init_profile(se_profile_t* profile, char* name) {
  profile->name = name;
  profile->call_count = INT64_C(0);
  profile->sources=NULL;
  profile->targets=NULL;
  profile->own.se_seconds=INT64_C(0);
  profile->own.se_nanoseconds=INT64_C(0);
  profile->cumul.se_seconds=INT64_C(0);
  profile->cumul.se_nanoseconds=INT64_C(0);
  profile->depth=INT64_C(0);
}

#define DRIFT_LOOP 1000

void get_profiler_started(se_local_profile_t* master) {

  se_profile_access_t* master_access;
  se_profile_t* master_profile = master->profile;

  /* Reckon the measure drift
   */
  se_time_t t1, t2;
  int i;
  current_time(&t1);
  for (i = 0; i < DRIFT_LOOP; i++){
    current_time(&t2);
  }
  subtract_time(&t1, &t2, &measure_drift);
  divide_time(&measure_drift, DRIFT_LOOP);
  current_time_counter = 0L;

  /* Prepare the "master" profile structure (not measured; just to have
   * uniform access to start_profile())
   */
  current_time(&(master->start));
  master_profile = &profiler_profile;
  master->profile = master_profile;
  master->cumul.se_seconds = INT64_C(0);
  master->cumul.se_nanoseconds = INT64_C(0);
  init_profile(master_profile, "<profiler>");
}

void start_profile(se_local_profile_t* source, se_local_profile_t* target) {

  se_profile_source_access_t* source_access;
  se_profile_target_access_t* target_access;
  se_profile_t* source_profile = source->profile;
  se_profile_t* target_profile = target->profile;
  int r;

  profile_nest++;
  target_profile->depth++;

  /* make sure the cumul counter is set to 0 */
  target->cumul.se_seconds = INT64_C(0);
  target->cumul.se_nanoseconds = INT64_C(0);

  /* now count that new link source->target in the source profile */
  source_access = profile_source_access(&(target_profile->sources), source_profile);
  source_access->count++;
  /* ... and in the target profile */
  target_access = profile_target_access(&(source_profile->targets), target_profile);
  target_access->count++;

  /* count the call */
  target_profile->call_count++;

  /* remember the number of times current_time() was called */
  target->current_time_counter = current_time_counter;

  /* set the start time for the target */
  current_time(&(target->start));
}

void stop_profile(se_local_profile_t* source, se_local_profile_t* target) {

  se_profile_t* target_profile;
  se_time_t stop, cumul, own, drift;
  int r;
  int current_time_counter_memory;
  se_profile_target_access_t* target_access;
  int out_of_recursive;

  /* set the stop time for the target */
  current_time(&stop);

  /* initialize variables */
  target_profile = target->profile;
  current_time_counter_memory = target->current_time_counter;
  target_access = profile_target_access(&(source->profile->targets), target_profile);
  target_profile->depth--;
  out_of_recursive = target_profile->depth == INT64_C(0);

  /* count the elapsed time
   *
   * cumul = stop - start
   *
   */
  r = subtract_time(&(target->start), &stop, &cumul);
  if (!r) fprintf(SE_ERR, "(stop-start < 0) %s <<< %s\n", source->profile->name, target->profile->name);

  /* take into account the measure drift
   *
   * cumul -= measure_drift * (number of times current_time() was called)
   *
   */
  drift = measure_drift;
  /*
    --*** CAD does not work! Why?

  r = multiply_time(&drift, current_time_counter - current_time_counter_memory);
  if (!r) fprintf(SE_ERR, "(could not calculate drift) %s <<< %s\n", source->profile->name, target->profile->name);
  */
  if (time_comparison(&cumul, &drift) > 0) {
    r = subtract_time(&drift, &cumul, &cumul);
    if (!r) fprintf(SE_ERR, "(could not subtract drift) %s <<< %s\n", source->profile->name, target->profile->name);
  }
  else {
    /*
     * It seems there is nothing we can do about it. It means that the
     * function took too small a time to run, and is under the
     * resolution of the timer. So be it: 0, then.
     */
    cumul.se_seconds = INT64_C(0);
    cumul.se_nanoseconds = INT64_C(0);
  }

  /* advertise the source
   *
   * local source cumul += cumul
   *
   */
  r = add_time(&(source->cumul), &cumul, &(source->cumul));
  if (!r) fprintf(SE_ERR, "(cannot add time to parent's cumul) %s <<< %s\n", source->profile->name, target->profile->name);

  /* now compute the target feature cumulative time
   *
   * global target cumul += cumul
   *
   */
  if (out_of_recursive) {
    r = add_time(&(target_profile->cumul), &cumul, &(target_profile->cumul));
    if (!r) fprintf(SE_ERR, "(cannot add cumul time to target) %s <<< %s\n", source->profile->name, target->profile->name);
  }

  /* ... and the target feature own time
   *
   * own = cumul - local target cumul
   * global target own += cumul
   *
   */
  r = subtract_time(&(target->cumul), &cumul, &own);
  if (!r) fprintf(SE_ERR, "(cumul less than time accumulated by children) %s <<< %s\n", source->profile->name, target->profile->name);
  r = add_time(&(target_profile->own), &own, &(target_profile->own));
  if (!r) fprintf(SE_ERR, "(cannot add own time to target) %s <<< %s\n", source->profile->name, target->profile->name);

  /* at last, reckon the cumul time on this access
   *
   * target_access->time += cumul
   */
  if (target_profile->depth <= INT64_C(1)) {
    r = add_time(&(target_access->time), &cumul, &(target_access->time));
    if (!r) fprintf(SE_ERR, "(cannot add cumul time to target access) %s <<< %s", source->profile->name, target->profile->name);
  }

  profile_nest--;
}

/* ====================================================================== */

/*
 * Finally, the functions below are used to generate the profile output.
 */

#define time_in_string(time, string, divisor) do { \
  sprintf((string), "%" PRId64 ".%.6" PRId64 " s", (time).se_seconds/(divisor), (((time).se_seconds%(divisor))*INT64_C(1000000) + (time).se_nanoseconds/INT64_C(1000))/(divisor)); \
} while(0)

void print_profile(FILE* file, se_profile_t* profile) {
  if (profile_nest) {
    /* It happens if the program is suddenly broken (^C or such) */
    fprintf(SE_ERR, "*** WARNING: profile_nest=%d\n", profile_nest);
    profile_nest=0;
  }

  if (profile->call_count > 0) {
    /*
     * We print the profile report only if the function was really called.
     */
    char own[32];
    char cumul[32];
    char avg_own[32];
    char avg_cumul[32];
    char target_time[32];
    se_profile_source_access_t* source = profile->sources;
    se_profile_source_access_t* source_head = source;
    se_profile_target_access_t* target = profile->targets;
    se_profile_target_access_t* target_head = target;
    time_in_string(profile->own,   own,   1);
    time_in_string(profile->cumul, cumul, 1);
    time_in_string(profile->own,   avg_own,   profile->call_count);
    time_in_string(profile->cumul, avg_cumul, profile->call_count);
    fprintf(file,
            "\n%s: %" PRId64 " call%s\n\tSelf time: %s (average: %s)\n\tCumulative time: %s (average: %s)\n",
            profile->name,
            profile->call_count,
            (profile->call_count > 1) ? "s" : "",
            own, avg_own,
            cumul, avg_cumul);
    if (source != NULL) {
      do {
        fprintf(file,
                "\tFrom %s: %" PRId64 " call%s (%.2f%%)\n",
                source->access->name,
                source->count,
                (source->count > 1) ? "s" : "",
                100.0 * source->count / profile->call_count);
        source=source->next;
      } while (source != source_head);
    }
    if (target != NULL) {
      do {
	time_in_string(target->time, target_time, 1);
        fprintf(file,
                "\tTo %s: %" PRId64 " call%s, %s\n",
                target->access->name,
                target->count,
                (target->count > 1) ? "s" : "",
		target_time);
        target=target->next;
      } while (target != target_head);
    }
  }
}

int profile_comparator(const void* left, const void* right) {
  /* We compare by cumulative time in order to have high level information */
  se_profile_t* left_profile = (se_profile_t*)left;
  se_profile_t* right_profile = (se_profile_t*)right;
  int r = time_comparison(&(left_profile->cumul), &(right_profile->cumul));
  if (r == 0) {
    r = (left_profile->call_count - right_profile->call_count);
  }
  return r;
}
