-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class ASSIGNMENT_HANDLER
	--
	-- Singleton object used to register all kinds of assignments from one type to another. First, one have to note
	-- that  assignments not only occurs in the assignment statement itself but also, for example, while passing
	-- some argument to a routine or for creation statements, or, as another example, for the assignment attempt.
	-- The `assignment_handler' singleton has two main goals. The most basic goal (historically speaking) is to
	-- provide IMPLICIT_CAST wrappers when they are necessary. The brand new usage of the `assignment_handler'
	-- (since the 1.0 SmartEiffel release) is to build the assignments type graph in order to compute more
	-- accurately the possible dynamic types of an expression (i.e. the corresponding RUN_TIME_SET). This
	-- information is important to implement dynamic dispatch and to perform the whole system covariance check
	-- (`safety_check').This singleton is shared via the GLOBALS.`assignment_handler' once function.
	--

inherit
	VISITABLE
insert
	GLOBALS

feature {ANY}
	accept (visitor: ASSIGNMENT_HANDLER_VISITOR) is
		do
			visitor.visit_assignment_handler(Current)
		end

	collect (site: POSITION; source, destination: TYPE) is
			-- Collect an actual assignment of `source' into `destination'. This allows the `assignment_handler'
			-- to compute the corresponding whole assignment graph and to prepare conversion functions (i.e.
			-- automatic conversion from expanded to reference). Finally, because of covariance, note that
			-- `source' is not always conforms to `destination'.
		require
			not_too_late: not smart_eiffel.status.is_adapting
			source /= Void
			destination /= Void
		local
			key, source_rtm: STRING
		do
			source_rtm := source.name.to_string
			if source_rtm = as_none then
			elseif source_rtm /= destination.name.to_string then
				if source.is_expanded then
					if destination.is_reference then
						key := collect_actual_conversion(source, destination)
						collect(site, source.actual_reference(destination), destination)
					end
				elseif destination.is_expanded then
					check
						source = destination.actual_reference(source)
					end
				else
					if destination.live_type = Void then
						smart_eiffel.collect_assignment(destination)
						--***PH(19/04/04) should be removed
					end
					if source.live_type = Void then
						smart_eiffel.collect_assignment(source)
						--***PH(19/04/04) should be removed
					end
					graph_node_for(source).assignment(site, destination)
				end
			end
		end

feature {ASSIGNMENT, EFFECTIVE_ARG_LIST, MANIFEST_ARRAY, PROC_CALL_1, FAKE_TUPLE, E_FUNCTION}
	implicit_cast (expression: EXPRESSION; expression_type, destination_type: TYPE): EXPRESSION is
			-- If necessary, wrap the source `expression' inside an IMPLICIT_CAST invisible wrapper object,
			-- hence the name of this function.
			-- Note: for most calls `implicit_cast' there is a corresponding `assignment' call during the
			-- collect step. (Usually called during specialize_2 or during inlining.)
		require
			smart_eiffel.status.is_collecting
			expression_type.is_a(destination_type, expression.start_position, expression.start_position)
		local
			cast: IMPLICIT_CAST
		do
			cast ?= expression
			if cast /= Void then
				Result := implicit_cast_(cast.expression, cast.expression_type, destination_type)
			else
				Result := implicit_cast_(expression, expression_type, destination_type)
			end
		ensure
			Result /= Void
		end

feature {}
	implicit_cast_ (expression: EXPRESSION; expression_type, destination_type: TYPE): EXPRESSION is
		local
			src_bit, dest_bit: TYPE_BIT
		do
			if expression_type = destination_type then
				Result := expression
			elseif expression_type.is_reference then
				if destination_type.is_reference then
					-- Reference to Reference:
					Result := expression
				else
					-- Reference to Expanded:
					create {IMPLICIT_CAST} Result.make(expression, expression_type, destination_type)
				end
			elseif destination_type.is_reference then
				-- Expanded to Reference:
				create {IMPLICIT_CAST} Result.make(expression, expression_type, destination_type)
			else
				check
					expression_type.is_expanded
					destination_type.is_expanded
				end
				if destination_type.is_integer and then expression_type.is_integer then
					-- INTEGER_* to INTEGER_*:
					create {IMPLICIT_CAST} Result.make(expression, expression_type, destination_type)
				elseif destination_type.is_real and then expression_type.is_integer then
					-- INTEGER to REAL:
					create {IMPLICIT_CAST} Result.make(expression, expression_type, destination_type)
				elseif destination_type.is_double then
					if expression_type.is_integer or else expression_type.is_real then
						-- INTEGER to DOUBLE or REAL to DOUBLE:
						create {IMPLICIT_CAST} Result.make(expression, expression_type, destination_type)
					else
						Result := expression
					end
				elseif destination_type.is_bit then
					src_bit ?= expression_type.canonical_type_mark
					dest_bit ?= destination_type.canonical_type_mark
					if src_bit.bit_count /= dest_bit.bit_count then
						-- BIT_* to BIT_*:
						create {IMPLICIT_CAST} Result.make(expression, expression_type, destination_type)
					else
						Result := expression
					end
				else
					error_handler.append("Internal error #3 in ASSIGNMENT_HANDLER from type ")
					error_handler.append(expression_type.name.to_string)
					error_handler.append(" to type ")
					error_handler.append(destination_type.name.to_string)
					error_handler.add_position(expression.start_position)
					error_handler.print_as_warning
					Result := expression
				end
			end
		end

feature {LIVE_TYPE}
	at_run_time (live_type: LIVE_TYPE; rts: RUN_TIME_SET) is
		require
			live_type = rts.owner
		local
			graph_node: GRAPH_NODE; unknown_position: POSITION
		do
			rts.add(unknown_position, live_type)
			graph_node := graph_node_for(live_type.type)
			graph_node.push_run_time_set(unknown_position, rts)
		end

feature {SMART_EIFFEL}
	reset is
			-- Called before a re-collect cycle.
		do
			collected_conversions.clear_count
			external_types.clear_count
		end

feature {IMPLICIT_CAST}
	collect_actual_conversion (source_type, destination_type: TYPE): STRING is
			-- To be called when we have to collect a true conversion form reference to the corresponding
			-- expanded or a true conversion from expanded to the corresponding reference.
		require
			source_type.is_expanded implies destination_type.is_reference
			source_type.is_reference implies destination_type.is_expanded
		local
			source, destination: TYPE; key: STRING
		do
			if source_type.is_expanded then
				source := source_type
				destination := source.actual_reference(destination)
			else
				destination := destination_type
				source := destination.actual_reference(source)
			end
			key := once "... unique buffer ..."
			key.clear_count
			key.extend('T')
			if source.live_type = Void then
				smart_eiffel.collect_cast(source)
			end
			source.live_type.id.append_in(key)
			key.append(once "toT")
			smart_eiffel.collect_cast(destination)
			destination.live_type.id.append_in(key)
			key := string_aliaser.string(key)
			if not collected_conversions.has(key) then
				collected_conversions.add([source, destination], key)
			end
			Result := key
		ensure
			string_aliaser.registered_one(Result)
		end

feature {AGENT_POOL}
	new_c_conversion_call (source, destination: TYPE): STRING is
			-- Prepare the beginning of the conversion function call.
			-- *** c_conversion_call should be removed *** (Dom. april 17th 2004) ***
		require
			source.live_type.at_run_time
			destination.live_type.at_run_time
		local
			key: STRING; i: INTEGER
		do
			if destination.is_separate then
				Result := once "(T0*)as_separate(self,"
			else
				key := collect_actual_conversion(source, destination)
				buffer.copy(key)
				buffer.extend('(')
				if source.is_reference then
					buffer.extend('(')
					from
						i := 1
					until
						key.item(i) = 't'
					loop
						buffer.extend(key.item(i))
						i := i + 1
					end
					buffer.append(once "*)")
				end
				Result := buffer
			end
		ensure
			not Result.is_empty
		end

feature {C_PRETTY_PRINTER, SWITCH}
	c_conversion_call (source, destination: TYPE) is
			-- *** use new_c_conversion_call instead *** (Dom. april 17th 2004) ***
		require
			source /= destination
		local
			key: STRING; i: INTEGER
		do
			if destination.is_separate then
				cpp.put_string(once "(T0*)as_separate(self,")
			else
				key := collect_actual_conversion(source, destination)
				cpp.put_string(key)
				cpp.put_character('(')
				if source.is_reference then
					cpp.put_character('(')
					from
						i := 1
					until
						key.item(i) = 't'
					loop
						cpp.put_character(key.item(i))
						i := i + 1
					end
					cpp.put_character('*')
					cpp.put_character(')')
				end
			end
		end

feature {C_PRETTY_PRINTER}
	c_definitions is
		local
			i: INTEGER; conversion: TUPLE[TYPE, TYPE]; key: STRING; source, destination: TYPE
		do
			from
				i := 1
			until
				i > collected_conversions.count
			loop
				key := collected_conversions.key(i)
				conversion := collected_conversions.item(i)
				source := conversion.first
				destination := conversion.second
				if source.is_bit then
					new_c_bit_bit_conversion(key, source, destination)
				elseif source.is_expanded and then destination.is_expanded then
					-- Because it is worthless or because this is already
					-- done by some macro definition (#define) in
					-- the "SmartEiffel/sys/runtime/base.h".
				else
					c_refexp_expref_conversion(key, source, destination)
				end
				i := i + 1
			end
		end

feature {SMART_EIFFEL}
	force_compatibility (reference_integer_set: SET[LIVE_TYPE]) is
		require
			reference_integer_set /= Void
		local
			i, j: INTEGER; t1, t2: TYPE; unknown_position: POSITION
		do
			from
				i := 1
			until
				i > reference_integer_set.count
			loop
				t1 := reference_integer_set.item(i).type
				from
					j := 1
				until
					j > reference_integer_set.count
				loop
					t2 := reference_integer_set.item(j).type
					if t1 /= t2 then
						if t1.is_a(t2, unknown_position, unknown_position) then
							collect(unknown_position, t1, t2)
						else
							error_handler.cancel
						end
					end
					j := j + 1
				end
				i := i + 1
			end
		end

	echo_information is
		local
			i, n: INTEGER; graph_node: GRAPH_NODE; rc: LIVE_TYPE; rts: RUN_TIME_SET
		do
			echo.put_string(once "Assignment graph: ")
			echo.put_integer(graph_nodes.count)
			from
				i := graph_nodes.lower
			until
				i > graph_nodes.upper
			loop
				graph_node := graph_nodes.item(i)
				rc := graph_node.live_type
				rts := rc.run_time_set
				n := n + graph_node.destination_types.count
				i := i + 1
			end
			echo.put_string(once " nodes and ")
			echo.put_integer(n)
			echo.put_string(once " transitions.%N")
		end

feature {INHERIT_TEXT_ITEM, RUN_TIME_SET}
	vncg (site: POSITION; source, destination: TYPE) is
		require
			source /= Void
			destination /= Void
		do
			if source.is_expanded then
				if destination.is_expanded then
				else
					collect(site, source, destination)
				end
			else
				check
					not destination.is_expanded
				end
				collect(site, source, destination)
			end
		end

feature {GRAPH_NODE}
	graph_node_for (type: TYPE): GRAPH_NODE is
		local
			key: HASHED_STRING
		do
			key := type.name
			Result := graph_nodes.reference_at(key)
			if Result = Void then
				create Result.make(type.canonical_type_mark)
				graph_nodes.put(Result, key)
			end
		end

feature {NATIVE, CECIL_ENTRY}
	from_external (type: TYPE; site: POSITION; args: FORMAL_ARG_LIST; result_type: TYPE_MARK) is
		require
			type /= Void
			not site.is_unknown
		local
			i: INTEGER
		do
			if args /= Void then
				from
					i := args.count
				until
					i <= 0
				loop
					from_external_(type, site, args.type_mark(i))
					i := i - 1
				end
			end
			if result_type /= Void then
				from_external_(type, site, result_type)
			end
		end

feature {}
	collected_conversions: DICTIONARY[TUPLE[TYPE, TYPE], STRING] is
			-- All needed conversions functions. The key is the corresponding
			-- name for the C function (e.g. T56toT78).
		once
			create {HASHED_DICTIONARY[TUPLE[TYPE, TYPE], STRING]} Result.make
		end

	c_header_for (key: STRING; source, destination: TYPE_MARK) is
			-- Compute the C header in the `buffer'
		do
			crash
			-- *** obsolete "use new_c_header_for"
			buffer.clear_count
			destination.c_type_for_result_in(buffer)
			buffer.extend(' ')
			buffer.append(key)
			buffer.extend('(')
			source.c_type_for_target_in(buffer)
			buffer.append(once " source)")
			echo.put_string(once "Conversion from %"")
			echo.put_string(source.written_mark)
			echo.put_string(once "%" to %"")
			echo.put_string(destination.written_mark)
			echo.put_string(once "%" is%Ndone by C function %"")
			echo.put_string(buffer)
			echo.put_character('%"')
			echo.put_character('%N')
		end

	new_c_header_for (key: STRING; source, destination: TYPE) is
			-- Compute the C header in the `buffer'
		do
			buffer.clear_count
			destination.c_type_for_result_in(buffer)
			buffer.extend(' ')
			buffer.append(key)
			buffer.extend('(')
			source.c_type_for_target_in(buffer)
			buffer.append(once " source)")
			echo.put_string(once "Conversion from %"")
			echo.put_string(source.name.to_string)
			echo.put_string(once "%" to %"")
			echo.put_string(destination.name.to_string)
			echo.put_string(once "%" is%Ndone by C function %"")
			echo.put_string(buffer)
			echo.put_character('%"')
			echo.put_character('%N')
		end

	c_refexp_expref_conversion (key: STRING; source, destination: TYPE) is
		require
			source /= destination
			source.live_type.at_run_time
			destination.live_type.at_run_time
			not (source.is_expanded and then destination.is_expanded)
		local
			source_lt, destination_lt: LIVE_TYPE; wa: ARRAY[RUN_FEATURE_2]; i, id: INTEGER; field: STRING
		do
			source_lt := source.live_type
			destination_lt := destination.live_type
			new_c_header_for(key, source, destination)
			cpp.put_c_heading(buffer)
			if destination.is_expanded then
				-- reference FOO to expanded FOO:
				check
					source.name.to_string.has_prefix(fz_reference) xor destination.name.to_string.has_prefix(fz_expanded)
				end
				if destination.is_kernel_expanded then
					if ace.no_check then
						cpp.put_string(once "vc((T0*)source,0);%N")
					end
					i := source_lt.run_time_set.count
					if i > 1 then
						cpp.put_string(once "switch(source->id){%N")
						from
						until
							i = 0
						loop
							id := source_lt.run_time_set.item(i).id
							cpp.put_string(once "case ")
							cpp.put_integer(id)
							cpp.put_string(once ": return")
							destination.mapping_cast
							cpp.put_string(once "(((T")
							cpp.put_integer(id)
							cpp.put_string(once "*)source)->_item);%N")
							i := i - 1
						end
						if ace.no_check then
							cpp.put_string(once "error0(%"")
							cpp.put_string(source_lt.name.to_string)
							cpp.put_string(once " expected (type error).%",NULL);%N")
						end
						cpp.put_string(once "}%N")
					else
						cpp.put_string(once "return((T")
						cpp.put_integer(source_lt.id)
						cpp.put_string(once "*)source)->_item;")
					end
				else
					cpp.put_character('T')
					cpp.put_integer(destination_lt.id)
					cpp.put_string(once " destination;%N")
					if ace.no_check then
						cpp.put_string(once "ci(")
						cpp.put_integer(source_lt.id)
						cpp.put_string(once ",(T0*)source,0);%N")
					end
					wa := destination_lt.writable_attributes
					if wa /= Void then
						from
							i := wa.lower
						until
							i > wa.upper
						loop
							field := wa.item(i).name.to_string
							cpp.put_string(once "destination._")
							cpp.put_string(field)
							cpp.put_string(once "=((T")
							cpp.put_integer(source_lt.id)
							cpp.put_string(once "*)source)->_")
							cpp.put_string(field)
							cpp.put_string(once ";%N")
							i := i + 1
						end
					end
					cpp.put_string(once "return destination;")
				end
			else
				-- expanded FOO to reference FOO :
				check
					source.is_expanded
				end
				cpp.put_character('T')
				cpp.put_integer(destination_lt.id)
				cpp.put_character('*')
				gc_handler.allocation_of(once "destination", destination_lt)
				if source.is_kernel_expanded then
					cpp.put_string(once "destination->_item=source;%N")
				else
					wa := destination_lt.writable_attributes
					if wa /= Void and then not source.is_dummy_expanded then
						from
							i := wa.lower
						until
							i > wa.upper
						loop
							field := wa.item(i).name.to_string
							if actually_has_the_corresponding_attribute_work_around(field, source_lt.writable_attributes) then
								cpp.put_string(once "destination->_")
								cpp.put_string(field)
								cpp.put_string(once "=source->_")
								cpp.put_string(field)
								cpp.put_string(once ";%N")
							end
							i := i + 1
						end
					end
				end
				cpp.put_string(once "return (T0*)destination;")
			end
			cpp.put_string(once "%N}%N")
		end

	c_bit_bit_conversion (key: STRING; source, destination: TYPE_MARK) is
		require
			source.is_static
			destination.is_static
			source.written_name /= destination.written_name
		local
			rc: LIVE_TYPE
		do
			crash
			-- obsolete "use new_c_bit_bit_conversion"
			c_header_for(key, source, destination)
			if destination.is_expanded then
				if source.is_expanded then
					-- expanded to expanded:
					buffer.copy(once "#define T")
					source.id.append_in(buffer)
					buffer.append(fz_to_t)
					destination.id.append_in(buffer)
					buffer.append(once "(x) (x)%N")
					cpp.put_string_on_h(buffer)
				else
					-- *_REF to expanded:
					cpp.put_c_heading(buffer)
					buffer.copy(once "return (((")
					source.c_type_for_target_in(buffer)
					buffer.append(once ")source)->_item);%N}%N")
					cpp.put_string(buffer)
				end
			else
				-- expanded to reference :
				cpp.put_c_heading(buffer)
				cpp.put_character('T')
				cpp.put_integer(destination.id)
				cpp.put_character('*')
				rc := destination.live_type
				gc_handler.allocation_of(once "destination", rc)
				if source.is_kernel_expanded then
					cpp.put_string(once "destination->_item=source;%N")
				else
					cpp.put_string(once "memcpy((((Tid*)destination)")
					if rc.is_tagged then
						cpp.put_string(once "+1")
					end
					cpp.put_string(once "),&source,sizeof(source));%N")
				end
				cpp.put_string(once "return ((T0*)destination);%N}%N")
			end
		end

	new_c_bit_bit_conversion (key: STRING; source, destination: TYPE) is
		require
			source /= destination
		local
			rc: LIVE_TYPE
		do
			new_c_header_for(key, source, destination)
			if destination.is_expanded then
				if source.is_expanded then
					-- expanded to expanded:
					buffer.copy(once "#define T")
					source.live_type.id.append_in(buffer)
					buffer.append(fz_to_t)
					destination.live_type.id.append_in(buffer)
					buffer.append(once "(x) (x)%N")
					cpp.put_string_on_h(buffer)
				else
					-- *_REF to expanded:
					cpp.put_c_heading(buffer)
					buffer.copy(once "return (((")
					source.c_type_for_target_in(buffer)
					buffer.append(once ")source)->_item);%N}%N")
					cpp.put_string(buffer)
				end
			else
				-- expanded to reference :
				cpp.put_c_heading(buffer)
				cpp.put_character('T')
				cpp.put_integer(destination.live_type.id)
				cpp.put_character('*')
				rc := destination.live_type
				gc_handler.allocation_of(once "destination", rc)
				if source.is_kernel_expanded then
					cpp.put_string(once "destination->_item=source;%N")
				else
					cpp.put_string(once "memcpy((((Tid*)destination)")
					if rc.is_tagged then
						cpp.put_string(once "+1")
					end
					cpp.put_string(once "),&source,sizeof(source));%N")
				end
				cpp.put_string(once "return ((T0*)destination);%N}%N")
			end
		end

	buffer: STRING is
		once
			create Result.make(128)
		end

	graph_nodes: DICTIONARY[GRAPH_NODE, HASHED_STRING] is
			-- The key STRING is the run time mark of the corresponding graph node.
		once
			create {HASHED_DICTIONARY[GRAPH_NODE, HASHED_STRING]} Result.with_capacity(512)
		end

	external_types: SET[TYPE] is
		once
			create {HASHED_SET[TYPE]} Result.make
		end

	from_external_ (type: TYPE; site: POSITION; type_mark: TYPE_MARK) is
		require
			not site.is_unknown
			type_mark /= Void
		local
			i: INTEGER; type1, type2: TYPE
		do
			type1 := type_mark.resolve_in(type)
			if not type1.is_expanded then
				if not external_types.has(type1) then
					from
						i := external_types.lower
					until
						i > external_types.count
					loop
						type2 := external_types.item(i) -- We must be pessimistic (i.e. everything can happens in the wild external world ;-):
						collect(site, type1, type2)
						collect(site, type2, type1)
						i := i + 1
					end
					external_types.add(type1)
				end
			end
		end

	actually_has_the_corresponding_attribute_work_around (field: STRING; wa: ARRAY[RUN_FEATURE_2]): BOOLEAN is
			-- ***
			-- This is clearly a work around because an expanded type should have exactly the same attributes than its
			-- corresponding reference type ... but I have the feeling that automatic conversion from expanded to reference will
			-- be removed of the langage, or I hope this will be removed.
			-- *** (Dom. june 30th 2004) ***
		local
			i: INTEGER
		do
			if wa /= Void then
				from
					i := wa.upper
				until
					Result or else i < wa.lower
				loop
					Result := wa.item(i).name.to_string = field
					i := i - 1
				end
			end
		end

	singleton_memory: ASSIGNMENT_HANDLER is
		once
			Result := Current
		end

invariant
	is_real_singleton: Current = singleton_memory

end -- class ASSIGNMENT_HANDLER
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
