-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class AGENT_CREATION
	--
	-- An Eiffel agent is a delayed call. Such an Eiffel object include a feature call with closed or open
	-- arguments (or closed or open target as well). An AGENT_CREATION object is the notation to create the
	-- agent (i.e. the feature is not launched yet, see also AGENT_EXPRESSION).
	--
	-- Examples of AGENT_CREATION:
	--          agent my_array.add_last(?)         -- The target is closed and the argument is open.
	--          agent {ARRAY[INTEGER]}.add_last(1) -- The target is open and the argument is closed.
	--          agent {ARRAY[INTEGER]}.add_last(?) -- Both the target and the argument are open.
	--          agent {ARRAY[INTEGER]}.add_last    -- A short-hand for the previous notation.
	--          agent my_array.put(?, 3)           -- Only the second argument is open.
	--          agent my_array.put(3, ?)           -- Only the first argument is open.
	--          agent {ARRAY[INTEGER]}.put         -- Target and all arguments are open.
	--

inherit
	EXPRESSION
		redefine is_equal
		end
	HASHABLE
		redefine is_equal
		end

creation {EIFFEL_PARSER}
	make

feature {ANY}
	start_position: POSITION
			-- Of the first letter of the mandatory "agent" keyword (obviously provided by the `eiffel_parser').

feature {AGENT_CREATION, AGENT_CREATION_VISITOR}
	base: TYPE_MARK
			-- Provided by the `eiffel_parser' when the target is open. This information is only memorized for
			-- pretty printing because this information is automatically wrapped by the `eiffel_parser' in a
			-- FAKE_TARGET of `base' type mark which is itself emmbedded in the `call' attribute.

	call: CALL
			-- All the information wrapped for convenience in a call object by the `eiffel_parser'. Actually,
			-- this call object as well as `base' and `start_position' are never modified during
			-- the whole compilation process.

	target: EXPRESSION
			-- Computed using `call'.

	feature_stamp: FEATURE_STAMP
			-- Computed using `target' and `call.feature_name'.

	arguments: EFFECTIVE_ARG_LIST
			-- Computed. May be Void when the feature to call has no arguments.

feature {AGENT_POOL, AGENT_CREATION}
	resolve_in_memory: TYPE
			-- Note: There is only one call left in AGENT_POOL for `resolve_in_memory' and it would be great to 
			-- be able to remove `resolve_in_memory' too.

feature {ANY}
	is_current, is_implicit_current: BOOLEAN is False

	extra_bracket_flag: BOOLEAN is False

	once_pre_computable: BOOLEAN is False

	hash_code: INTEGER is
		do
			Result := call.feature_name.hash_code
		end

	is_equal (other: like Current): BOOLEAN is
		do
			check
				feature_stamp /= Void
				other.feature_stamp /= Void
			end
			Result := target = other.target and then feature_stamp = other.feature_stamp and then arguments = other.arguments and then resolve_in_memory = other.resolve_in_memory
		end

	declaration_type: TYPE is
		local
			tlc, dt: TYPE; fs: FEATURE_STAMP
		do
			tlc := start_position.class_text.declaration_type_of_like_current
			dt := target.declaration_type
			fs := dt.lookup(call.feature_name)
			Result := compute_agent_type(tlc, dt, fs)
		end

	specialize_in (type: TYPE): like Current is
		local
			t: like target; a: like arguments
		do
			t := target.specialize_in(type)
			a := arguments
			if a /= Void then
				a := a.specialize_in(type)
			end
			Result := current_or_twin_init(Void, t, feature_stamp, a)
		end

	specialize_thru (parent_type: TYPE; inherit_item: INHERIT_TEXT_ITEM; new_type: TYPE): like Current is
		local
			t: like target; a: like arguments
		do
			t := target.specialize_thru(parent_type, inherit_item, new_type)
			a := arguments
			if a /= Void then
				a := a.specialize_thru(parent_type, inherit_item, new_type)
			end
			Result := current_or_twin_init(Void, t, feature_stamp, a)
		end

	specialize_2 (type: TYPE): like Current is
		local
			t: like target; a: like arguments; fs: FEATURE_STAMP; af: ANONYMOUS_FEATURE; target_type: TYPE
		do
			t := target.specialize_2(type)
			fs := t.declaration_type.search(call.feature_name)
			if fs = Void then
				smart_eiffel.unknown_feature_fatal_error(target, t.declaration_type, call.feature_name)
			end
			target_type := t.resolve_in(type)
			fs := fs.resolve_dynamic_binding_for(target_type)
			if feature_stamp = Void then
				feature_stamp := fs
			end
			af := fs.anonymous_feature(target_type)
			if af.arguments /= Void then
				if arguments = Void then
					arguments := af.arguments.omitted_open_arguments(call.feature_name.start_position)
				end
				open_args_buffer.clear_count
				a := arguments.notify_open_operands(open_args_buffer, target_type, af)
				open_args := open_args_buffer.twin
				a := a.specialize_2(type, af, target_type, target.is_current)
			end
			smart_eiffel.argument_count_check(start_position, af, arguments)
			Result := current_or_twin_init(type, t, fs, a)
		end

	has_been_specialized: BOOLEAN is
		do
			Result := target.has_been_specialized
			if Result and then arguments /= Void then
				Result := arguments.has_been_specialized
			end
		end

	resolve_in (type: TYPE): TYPE is
		local
			target_type: TYPE
		do
			target_type := target.resolve_in(type)
			check
				target.declaration_type.lookup(call.feature_name).resolve_dynamic_binding_for(target_type)
				= feature_stamp
			end
			Result := compute_agent_type(type, target_type, feature_stamp)
		end

	collect (type: TYPE): TYPE is
		local
			target_type, unused: TYPE
		do
			target_type := target.collect(type)
			unused := smart_eiffel.collect(target_type, feature_stamp, False)
			if arguments /= Void then
				arguments.collect(type, feature_stamp, target_type)
			end
			Result := compute_agent_type(type, target_type, feature_stamp)
			smart_eiffel.collect_create(Result)
			agent_pool.agent_creation_collect(Current, Result)
		end

	adapt_for (type: TYPE): like Current is
		local
			t: like target; a: like arguments
		do
			t := target.adapt_for(type)
			if arguments /= Void then
				a := arguments.adapt_for(type)
			end
			Result := current_or_twin_init(type, t, feature_stamp, a)
		end

	non_void_no_dispatch_type (type: TYPE): TYPE is
		do
		end

	simplify (type: TYPE): EXPRESSION is
		local
			t: like target; a: like arguments
		do
			t := target.simplify(type)
			a := arguments
			if a /= Void then
				a := a.simplify(type)
			end
			Result := current_or_twin_init(type, t, feature_stamp, a)
		end

	accept (visitor: AGENT_CREATION_VISITOR) is
		do
			visitor.visit_agent_creation(Current)
		end

	is_void: BOOLEAN is False

	is_writable: BOOLEAN is False

	is_manifest_string: BOOLEAN is False

	is_static: BOOLEAN is False

	is_result: BOOLEAN is False

	side_effect_free (type: TYPE): BOOLEAN is
		do
			-- Memory allocation.
		end

	safety_check is
		do
			target.safety_check
			if arguments /= Void then
				arguments.safety_check
			end
			-- *** There is probably something else to check ???
			-- (Dom. feb 7th 2004)
			-- ***
		end

	compile_to_c (type: TYPE) is
		local
			mold_id: STRING; i: INTEGER; boost, need_comma: BOOLEAN; e: EXPRESSION
		do
			mold_id := once "......................"
			mold_id.clear_count
			mold_id_in(type, mold_id)
			agent_pool.register_for_c_define(mold_id, Current)
			cpp.put_string(mold_id)
			cpp.put_character('(')
			boost := ace.boost
			if not boost then
				cpp.put_string(once "&ds")
				need_comma := True
			end
			if base = Void then
				if need_comma then
					cpp.put_character(',')
				end
				target.mapping_c_arg(type, target.resolve_in(type).canonical_type_mark)
				need_comma := True
			end
			if arguments /= Void then
				from
					i := 1
				until
					i > arguments.count
				loop
					if is_closed_argument(i) then
						if need_comma then
							cpp.put_character(',')
						end
						e := arguments.expression(i)
						e.mapping_c_arg(type, e.resolve_in(type).canonical_type_mark)
						need_comma := True
					end
					i := i + 1
				end
			end
			cpp.put_character(')')
		end

	mapping_c_target (type: TYPE; formal_type: TYPE_MARK) is
		do
			compile_to_c(type)
		end

	mapping_c_arg (type: TYPE; formal_arg_type: TYPE_MARK) is
		do
			compile_to_c(type)
		end

	use_current: BOOLEAN is
		do
			if target.use_current then
				Result := True
			elseif arguments /= Void then
				Result := arguments.use_current
			end
		end

	bracketed_pretty, pretty (indent_level: INTEGER) is
		do
			pretty_printer.keyword(once "agent")
			if base /= Void then
				pretty_printer.put_character('{')
				pretty_printer.put_type_mark(base)
				pretty_printer.put_character('}')
				pretty_printer.put_character('.')
			end
			call.pretty(indent_level)
		end

	pretty_target (indent_level: INTEGER) is
		do
			pretty_printer.put_character('(')
			pretty(indent_level)
			pretty_printer.put_character(')')
			pretty_printer.put_character('.')
		end

	short (type: TYPE) is
		do
			short_printer.hook_or(as_agent, as_agent)
			short_printer.put_character(' ')
			if base /= Void then
				short_printer.hook_or(once "open_curly_bracket", once "{")
				base.short(type)
				short_printer.hook_or(once "close_curly_bracket", once "}")
				short_printer.hook_or(once "dot", once ".")
			end
			call.feature_name.short(type)
			if call.arguments /= Void then
				call.arguments.short(type)
			end
		end

	short_target (type: TYPE) is
		do
			bracketed_short(type)
			short_printer.put_dot
		end

	precedence: INTEGER is
		do
			Result := atomic_precedence
		end

	jvm_assign_creation, jvm_assign (type: TYPE) is
		do
			check
				False
			end
		end

	compile_to_jvm_into (type: TYPE; dest: TYPE_MARK): INTEGER is
		do
			Result := 1
			compile_to_jvm(type)
		end

	compile_target_to_jvm, compile_to_jvm (type: TYPE) is
		do
			not_yet_implemented
		end

	jvm_branch_if_false, jvm_branch_if_true (type: TYPE): INTEGER is
		do
			not_yet_implemented
		end

feature {AGENT_POOL}
	mold_id_in (type: TYPE; buffer: STRING) is
			-- Identify the creation form (which feature is called and where are placed base / open arguments).
		local
			target_type: TYPE; open_operand: OPEN_OPERAND; i: INTEGER; args: EFFECTIVE_ARG_LIST
			fake_target: FAKE_TARGET; expression: EXPRESSION
		do
			-- Do _NOT_ change the following mangling because it is used by `mold_id_is_a' in AGENT_POOL.
			buffer.append(once "agent")
			target_type := target.resolve_in(type)
			target_type.live_type.id.append_in(buffer)
			buffer.append(call.feature_name.to_string)
			if feature_stamp.anonymous_feature(target_type).result_type = Void then
				buffer.append(once "_P_")
			else
				buffer.append(once "_F_")
			end
			fake_target ?= target
			agent_pool.open_close_mold_id_in(buffer, fake_target /= Void, target_type)
			args := arguments
			if args /= Void then
				from
					i := 1
				until
					i > args.count
				loop
					expression := args.expression(i)
					open_operand ?= expression
					agent_pool.open_close_mold_id_in(buffer, open_operand /= Void, expression.resolve_in(type))
					i := i + 1
				end
			end
		end
	
	c_define_function (type: TYPE) is
		require
			cpp.on_c
		local
			boost: BOOLEAN; target_type, at, agent_result: TYPE; tm: TYPE_MARK; mold_id_index, i: INTEGER
			mold_id, buffer: STRING; rf: RUN_FEATURE
		do
			boost := ace.boost
			mold_id := once "......................"
			mold_id.clear_count
			mold_id_in(type, mold_id)
			mold_id_index := agent_pool.mold_id_index_of(mold_id)
			cpp.swap_on_h
			c_define_type(type, boost, mold_id, mold_id_index)
			cpp.swap_on_c
			at := resolve_in(type)
			agent_result := at.agent_result
			-- The local execution function:
			buffer := once "............................ unique buffer .........................."
			buffer.copy(once "%Nstatic ")
			if agent_result /= Void then
				agent_result.canonical_type_mark.c_type_for_result_in(buffer)
			else
				buffer.append(fz_void)
			end
			buffer.append(once " _")
			buffer.append(mold_id)
			buffer.extend('(')
			if not boost then
				buffer.append(once "se_dump_stack*caller,")
			end
			buffer.append(once "se_")
			buffer.append(mold_id)
			buffer.append(once "*u")
			target_type := target.resolve_in(type)
			if base /= Void then
				buffer.extend(',')
				target_type.canonical_type_mark.c_type_for_result_in(buffer)
				buffer.extend(' ')
				buffer.extend('C')
			end
			if open_args /= Void then
				from
					i := 1
				until
					i > arguments.count
				loop
					if not is_closed_argument(i) then
						buffer.extend(',')
						tm := arguments.expression(i).resolve_in(type).canonical_type_mark
						tm.c_type_for_result_in(buffer)
						buffer.extend(' ')
						buffer.extend('a')
						i.append_in(buffer)
					end
					i := i + 1
				end
			end
			buffer.append(once "){%N")
			if base = Void then
				tm := target.resolve_in(type).canonical_type_mark
				if tm.is_user_expanded then
					tm.c_type_for_target_in(buffer)
					buffer.append(once " C=&(u->C);%N")
				else
					tm.c_type_for_argument_in(buffer)
					buffer.append(once " C=u->C;%N")
				end
			end
			if arguments /= Void then
				from
					i := 1
				until
					i > arguments.count
				loop
					if is_closed_argument(i) then
						tm := arguments.expression(i).resolve_in(type).canonical_type_mark
						tm.c_type_for_argument_in(buffer)
						buffer.extend(' ')
						buffer.extend('a')
						i.append_in(buffer)
						buffer.append(once "=u->a")
						i.append_in(buffer)
						buffer.extend(';')
						buffer.extend('%N')
					end
					i := i + 1
				end
			end
			cpp.put_string(buffer)
			if not boost then
				cpp.put_string(once "[
                 se_frame_descriptor fd={"<agent-call-wrapper-2>",0,0,"",1};
                 se_dump_stack ds;
                 ds.fd=&fd;
                 ds.p=0;
                 ds.caller=caller;
                 ds.exception_origin=NULL;
                 ds.locals=NULL;

                  ]")
			end
			-- Calling the actual one:
			if agent_result /= Void then
				cpp.put_string(once "u->R=")
			end
			rf := feature_stamp.run_feature_for(target_type)
			cpp.push_pop_agent_launcher(rf.type_of_current, rf)
			if agent_result /= Void then
				cpp.put_string(once ";%Nreturn u->R;%N")
			end
			cpp.put_character('}')
			-- The definition function:
			buffer.copy(once "/*agent*/T0*")
			buffer.append(mold_id)
			buffer.extend('(')
			if not boost then
				buffer.append(once "se_dump_stack*caller")
			end
			if base = Void then
				if buffer.last /= '(' then
					buffer.extend(',')
				end
				tm := target.resolve_in(type).canonical_type_mark
				tm.c_type_for_argument_in(buffer)
				buffer.append(once " C")
			end
			if arguments /= Void then
				from
					i := 1
				until
					i > arguments.count
				loop
					if is_closed_argument(i) then
						if buffer.last /= '(' then
							buffer.extend(',')
						end
						tm := arguments.expression(i).resolve_in(type).canonical_type_mark
						tm.c_type_for_argument_in(buffer)
						buffer.extend(' ')
						buffer.extend('a')
						i.append_in(buffer)
					end
					i := i + 1
				end
			end
			if buffer.last = '(' then
				buffer.append(fz_void)
			end
			buffer.extend(')')
			cpp.put_c_heading(buffer)
			cpp.put_string(once "se_")
			cpp.put_string(mold_id)
			cpp.put_string(once "*u=(void*)new_agent(")
			cpp.put_integer(at.live_type.id)
			cpp.put_string(once ");%Nu->creation_mold_id=")
			cpp.put_integer(mold_id_index)
			cpp.put_string(once ";%Nu->afp=_")
			cpp.put_string(mold_id)
			cpp.put_string(once ";%N#ifdef SE_GC_LIB%N%
			     %u->gc_mark_agent_mold=gc_mark_")
			cpp.put_string(mold_id)
			cpp.put_string(once ";%N#endif%N")
			if base = Void then
				if ace.no_check and then not target.resolve_in(type).is_expanded and then not target.is_current then
					cpp.put_string(once "u->C=vc(C,")
					cpp.put_position(target.start_position)
					cpp.put_string(once ");%N")
				else
					cpp.put_string(once "u->C=C;%N")
				end
			end
			if arguments /= Void then
				from
					i := 1
				until
					i > arguments.count
				loop
					if is_closed_argument(i) then
						cpp.put_string(once "u->a")
						cpp.put_integer(i)
						cpp.put_string(once "=a")
						cpp.put_integer(i)
						cpp.put_string(once ";%N")
					end
					i := i + 1
				end
			end
			cpp.put_string(once "return((T0*)u);%N}%N")
			-- The gc_mark_MOLD_ID function:
			cpp.put_string(once "%N#ifdef SE_GC_LIB%N")
			buffer.copy(once "void gc_mark_")
			buffer.append(mold_id)
			buffer.append(once "(se_")
			buffer.append(mold_id)
			buffer.append(once "*u)")
			cpp.put_string_on_h(once "%N#ifdef SE_GC_LIB%N")
			cpp.put_c_heading(buffer)
			cpp.put_string_on_h(once "#endif%N%N")
			buffer.copy(once "gc_agent*gcu=(gc_agent*)u;%N%
			    %if (gcu->header.flag==FSOH_UNMARKED){%N%
			    %gcu->header.flag=FSOH_MARKED;%N")
			if base = Void then
				tm := target.resolve_in(type).canonical_type_mark
				if tm.is_reference then
					buffer.append(once "gc_mark(u->C);%N")
				else
					check
						tm.is_expanded
					end
					if tm.need_gc_mark_function then
						tm.gc_mark_in(buffer)
						buffer.append(once "(&(u->C));%N")
					end
				end
			end
			if arguments /= Void then
				from
					i := 1
				until
					i > arguments.count
				loop
					if is_closed_argument(i) then
						tm := arguments.expression(i).resolve_in(type).canonical_type_mark
						if tm.is_reference then
							buffer.append(once "gc_mark(u->a")
							i.append_in(buffer)
							buffer.append(once ");%N")
						else
							check
								tm.is_expanded
							end
							if tm.need_gc_mark_function then
								tm.gc_mark_in(buffer)
								buffer.append(once "(&(u->a")
								i.append_in(buffer)
								buffer.append(once "));%N")
							end
						end
					end
					i := i + 1
				end
			end
			buffer.append(once "}%N}%N#endif%N")
			cpp.put_string(buffer)
		ensure
			cpp.on_c
		end

feature {AGENT_CREATION}
	init (rim: like resolve_in_memory; t: like target; fs: like feature_stamp; a: like arguments) is
		require
			t /= Void
		do
			resolve_in_memory := rim
			target := t
			arguments := a
			feature_stamp := fs
		ensure
			resolve_in_memory = rim
			target = t
			arguments = a
			feature_stamp = fs
		end

feature {}
	open_args: FAST_ARRAY[OPEN_OPERAND]
			-- Computed using `arguments' to store, if any, the list of open arguments. -- (The `target' is never
			-- part of this list, see `base'.)

	open_tuple: TUPLE_TYPE_MARK

	is_closed_argument (i: INTEGER): BOOLEAN is
			-- Is the `i'th argument a closed argument?
		require
			i.in_range(1, arguments.count)
		local
			e: EXPRESSION; open_operand: OPEN_OPERAND
		do
			e := arguments.expression(i)
			open_operand ?= e
			Result := open_operand = Void
		end

	compute_agent_type (type: TYPE; target_type: TYPE; fs: FEATURE_STAMP): TYPE is
			-- *** Still leaky and slow at time beeing (Dom. feb 7th 2004) ***
		require
			type /= Void
			fs.has_anonymous_feature_for(target_type)
		local
			actual_base, result_tm: TYPE_MARK; open_type_list: ARRAY[TYPE_MARK]; i: INTEGER
			af: ANONYMOUS_FEATURE; sp: POSITION
		do
			sp := type.class_text.name.start_position
			actual_base := target_type.canonical_type_mark
			if base /= Void then
				create open_type_list.with_capacity(4, 1)
				open_type_list.add_last(base.to_static(type))
			end
			if open_args /= Void then
				from
					i := open_args.lower
				until
					i > open_args.upper
				loop
					if open_type_list = Void then
						create open_type_list.with_capacity(4, 1)
					end
					open_type_list.add_last(open_args.item(i).resolve_in(type).canonical_type_mark)
					i := i + 1
				end
			end
			if open_type_list /= Void then
				create {NON_EMPTY_TUPLE_TYPE_MARK} open_tuple.make(sp, open_type_list)
			else
				create {EMPTY_TUPLE_TYPE_MARK} open_tuple.make(sp)
			end
			check
				open_tuple.is_static
			end
			af := fs.anonymous_feature(target_type)
			result_tm := af.result_type
			if result_tm = Void then
				create {AGENT_TYPE_MARK} result_tm.procedure(sp, actual_base, open_tuple.as_type_mark)
			else
				result_tm := result_tm.to_static(type)
				if result_tm.is_boolean then
					create {AGENT_TYPE_MARK} result_tm.predicate(sp, actual_base, open_tuple.as_type_mark)
				else
					create {AGENT_TYPE_MARK} result_tm.function(sp, actual_base, open_tuple.as_type_mark, result_tm)
				end
			end
			check
				result_tm.is_static
			end
			Result := result_tm.type
		ensure
			Result /= Void
		end

	make (sp: like start_position; b: like base; c: like call) is
		require
			not sp.is_unknown
			c /= Void
		do
			start_position := sp
			base := b
			call := c
			target := c.target
			arguments := c.arguments
		ensure
			start_position = sp
			base = b
			call = c
		end

	current_or_twin_init (rim: like resolve_in_memory; t: like target; fs: like feature_stamp; a: like arguments): like Current is
		require
			t /= Void
		do
			if t = target and then a = arguments and then fs = feature_stamp then
				if resolve_in_memory = Void or else resolve_in_memory = rim then
					resolve_in_memory := rim
					Result := Current
				else
					Result := twin
					Result.init(rim, t, fs, a)
				end
			else
				Result := twin
				Result.init(rim, t, fs, a)
			end
		ensure
			Result.target = t
			Result.feature_stamp = fs
			Result.arguments = a
		end

	open_args_buffer: FAST_ARRAY[OPEN_OPERAND] is
		once
			create Result.with_capacity(16)
		end

	c_define_type (type: TYPE; boost: BOOLEAN; mold_id: STRING; mold_id_index: INTEGER) is
		require
			cpp.on_h
			boost = ace.boost
		local
			i: INTEGER; at, t, target_type: TYPE; tm: TYPE_MARK; open_arguments: ARRAY[TYPE]; buffer: STRING
		do
			buffer := once "............................ unique buffer .........................."
			buffer.copy(once "typedef struct _se_")
			buffer.append(mold_id)
			buffer.append(once " se_")
			buffer.append(mold_id)
			buffer.append(once ";%Nstruct _se_")
			buffer.append(mold_id)
			buffer.append(once "{Tid id;%Nint creation_mold_id;%N")
			at := resolve_in(type)
			t := at.agent_result
			if t = Void then
				buffer.append(fz_void)
			else
				t.canonical_type_mark.c_type_for_result_in(buffer)
			end
			buffer.append(once "(*afp)(")
			if not boost then
				buffer.append(once "se_dump_stack*,")
			end
			buffer.append(once "se_")
			buffer.append(mold_id)
			buffer.extend('*')
			open_arguments := at.open_arguments
			if open_arguments /= Void then
				from
					i := open_arguments.lower
				until
					i > open_arguments.upper
				loop
					buffer.extend(',')
					tm := open_arguments.item(i).canonical_type_mark
					tm.c_type_for_result_in(buffer)
					i := i + 1
				end
			end
			buffer.append(once "[
			     );
			     void(*gc_mark_agent_mold)(se_
			     ]")
			buffer.append(mold_id)
			buffer.append(once "*);%N")
			if base = Void then
				target_type := target.resolve_in(type)
				target_type.canonical_type_mark.c_type_for_argument_in(buffer)
				buffer.append(once " C;")
			end
			if arguments /= Void then
				from
					i := 1
				until
					i > arguments.count
				loop
					if is_closed_argument(i) then
						tm := arguments.expression(i).resolve_in(type).canonical_type_mark
						tm.c_type_for_argument_in(buffer)
						buffer.append(once " a")
						i.append_in(buffer)
						buffer.extend(';')
					end
					i := i + 1
				end
			end
			t := at.agent_result
			if t /= Void then
				t.canonical_type_mark.c_type_for_argument_in(buffer)
				buffer.append(once " R;")
			end
			buffer.append(once "};%N")
			cpp.put_string(buffer)
		ensure
			cpp.on_h
		end

invariant
	not start_position.is_unknown

	call /= Void

	target /= Void

end -- class AGENT_CREATION
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
