-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class FEATURE_NAME
	--
	-- For all kinds of feature names (simple names, infix or prefix names, frozen names as well as Precursor
	-- names).
	-- Note 1: the same class is used for all kinds of names to avoid polymorphism in order to be able to use this
	-- class as en entry for some DICTIONARY or SET as well.
	-- Note 2: this class is also used as an EXPRESSION for attribute when they are syntactically recognizable
	-- (i.e. left hand side of an assignment or for example, the target of some creation instruction).
	--

inherit
	COMPARABLE
		redefine is_equal
		end
	HASHABLE
		redefine is_equal
		end
	EXPRESSION
		redefine is_equal
		end
insert
	NAME
		redefine is_equal
		end

creation {ANY}
	ordinary_name, simple_feature_name, infix_name, prefix_name, frozen_name, unknown_position

feature {ANY}
	start_position: POSITION
			-- Of the first character.

	name: HASHED_STRING
			-- The unique corresponding one.

	key_name: HASHED_STRING
			-- To avoid clash between different kinds of names (for example when using same infix / prefix
			-- operator). Also used to compute the C name or the JVM name of a feature.

	hash_code: INTEGER
			-- To cache `key_name.hash_code'.

	feature_stamp: FEATURE_STAMP
			-- The good feature_stamp for the feature in some type.

	to_string: STRING is
			-- To cache `name.to_string'.
		do
			Result := name.to_string
		end

	to_key: STRING is
			-- To cache `key_name.to_string'.
		do
			Result := key_name.to_string
		end

	is_equal (other: like Current): BOOLEAN is
		do
			Result := key_name = other.key_name
		end

	infix "<" (other: like Current): BOOLEAN is
		do
			Result := to_key < other.to_key
		end

	is_frozen: BOOLEAN is
		do
			Result := name_type = C_frozen_name
		end

	is_simple_feature_name: BOOLEAN is
		do
			Result := name_type = C_simple_feature_name
		end

	is_infix_name: BOOLEAN is
		do
			Result := name_type = C_infix_name
		end

	is_prefix_name: BOOLEAN is
		do
			Result := name_type = C_prefix_name
		end

	is_current: BOOLEAN is False

	is_implicit_current: BOOLEAN is False

	is_manifest_string: BOOLEAN is False

	is_result: BOOLEAN is False

	is_void: BOOLEAN is False

	side_effect_free (type: TYPE): BOOLEAN is
		do
			Result := True
		end

	once_pre_computable: BOOLEAN is False

	precedence: INTEGER is
		do
			Result := atomic_precedence
		end

	use_current: BOOLEAN is True

	is_writable: BOOLEAN is
		do
			Result := C_simple_feature_name = name_type
		end

	infix_or_prefix: STRING is
			-- Return "infix" or "prefix" or something appropriate.
		do
			inspect
				name_type
			when C_simple_feature_name then
			when C_infix_name then
				Result := fz_infix
			when C_prefix_name then
				Result := fz_prefix
			when C_frozen_name then
				Result := feature_name.infix_or_prefix
			end
		end

	complete_name: STRING is
		do
			Result := complete_name_memory
			if Result = Void then
				Result := ""
				complete_name_memory := Result
				inspect
					name_type
				when C_simple_feature_name then
					Result.append(to_string)
				when C_infix_name then
					Result.append(once "infix %"")
					Result.append(to_string)
					Result.extend('%"')
				when C_prefix_name then
					Result.append(once "prefix %"")
					Result.append(to_string)
					Result.extend('%"')
				when C_frozen_name then
					Result.copy(feature_name.complete_name)
				end
			end
		ensure
			Result /= Void
		end

	mapping_c_in (str: STRING) is
		do
			str.append(to_key)
		end

	result_type: TYPE_MARK is
		do
			check
				C_simple_feature_name = name_type
			end
			Result := run_feature_2.result_type
		end

	specialize_in (type: TYPE): like Current is
		local
			fs: like feature_stamp; type_mark: TYPE_MARK; dt: like declaration_type
		do
			fs := type.search(Current)
			if fs = Void then
				error_handler.add_position(start_position)
				error_handler.append("Attribute ")
				error_handler.add_expression(Current)
				error_handler.append(" not found in type ")
				error_handler.add_type(type)
				error_handler.append(".")
				error_handler.print_as_fatal_error
			end
			if feature_stamp = Void then
				feature_stamp := fs
			end
			type_mark := type.feature_accumulator.find_type_for(fs)
			if type_mark = Void then
				error_handler.add_position(start_position)
				error_handler.append("This feature name is not an expression (no result and not writable).")
				error_handler.print_as_fatal_error
			end
			dt := type_mark.resolve_in(type)
			if declaration_type = Void then
				declaration_type := dt
			end
			if feature_stamp = fs and then declaration_type = dt then
				Result := Current
			else
				Result := twin
				Result.set_feature_stamp_and_type(fs, dt)
			end
		end

	specialize_thru (parent_type: TYPE; inherit_item: INHERIT_TEXT_ITEM; new_type: TYPE): like Current is
		local
			fs: like feature_stamp
		do
			check
				feature_stamp /= Void
			end
			fs := feature_stamp.specialize_thru(parent_type, inherit_item, new_type)
			if fs /= feature_stamp then
				Result := twin
				Result.set_feature_stamp(fs)
			else
				Result := Current
			end
		end

	specialize_2 (type: TYPE): like Current is
		local
			anonymous_feature: ANONYMOUS_FEATURE; writable_attribute: WRITABLE_ATTRIBUTE
		do
			anonymous_feature := feature_stamp.anonymous_feature(type)
			writable_attribute ?= anonymous_feature
			if writable_attribute = Void then
				error_handler.add_position(start_position)
				error_handler.append("Feature found is not writable (i.e. not an attribute).")
				error_handler.print_as_error
				if anonymous_feature = Void then
					error_handler.print_as_fatal_error
				end
				error_handler.append(" See also the next message for more information.")
				error_handler.print_as_error
				error_handler.add_position(anonymous_feature.start_position)
				error_handler.append("Here is the corresponding feature definition (not an attribute).")
				error_handler.print_as_fatal_error
			end
			Result := Current
		end

	has_been_specialized: BOOLEAN is
		do
			Result := feature_stamp /= Void and then declaration_type /= Void
		end

	adapt_for (t: TYPE): like Current is
		local
			wbc: CLASS_TEXT; rf2: like run_feature_2; ct: TYPE_MARK
		do
			check
				C_simple_feature_name = name_type
			end
			ct := t.canonical_type_mark
			wbc := start_position.class_text
			rf2 ?= feature_stamp.run_feature_for(t) -- *** ?? t.get_feature_name(feature_stamp))
			check
				rf2 /= Void
			end
			if run_feature_2 = Void then
				run_feature_2 := rf2
				Result := Current
			elseif run_feature_2 = rf2 then
				Result := Current
			else
				Result := twin
				Result.set_run_feature_2(rf2)
			end
		end

	non_void_no_dispatch_type (type: TYPE): TYPE is
		do
			-- There is probably better to do here, but actually, such
			-- object are not so common.
		end

	simplify (type: TYPE): EXPRESSION is
		do
			Result := Current
		end

	resolve_in (type: TYPE): TYPE is
		local
			tm: TYPE_MARK
		do
			tm := feature_stamp.anonymous_feature(type).result_type -- If this call fail, then check if specialize_2 has been called!
			if tm /= Void then
				Result := tm.resolve_in(type)
			else
				error_handler.append("This feature is a procedure: it cannot be used as target.")
				error_handler.add_position(feature_stamp.anonymous_feature(type).start_position)
				error_handler.add_position(start_position)
				error_handler.print_as_fatal_error
			end
		end

	declaration_type: TYPE

	safety_check is
		do
		end

	compile_to_c (type: TYPE) is
		do
			check
				C_simple_feature_name = name_type
			end
			cpp.put_string(once "/*SFN*/")
			if smart_eiffel.scoop and then run_feature_2.current_type.is_separate then
				cpp.put_string(once "(((T")
				cpp.put_integer(run_feature_2.current_type.id)
				cpp.put_string(once "*)C)->ref->")
			else
				cpp.put_string(once "(C->")
			end
			run_feature_2.put_c_field_name
			cpp.put_character(')')
		end

	mapping_c_target (type: TYPE; target_type: TYPE_MARK) is
		local
			flag: BOOLEAN
		do
			check
				C_simple_feature_name = name_type
			end
			flag := cpp.call_invariant_start(target_type)
			compile_to_c(type)
			if flag then
				cpp.call_invariant_end
			end
		end

	mapping_c_arg (type: TYPE; formal_arg_type: TYPE_MARK) is
		do
			check
				C_simple_feature_name = name_type
			end
			compile_to_c(type)
		end

	is_static: BOOLEAN is False

	origin_base_class: CLASS_TEXT is
			-- Void or the CLASS_TEXT where `Current' is written in.
		do
			Result := start_position.class_text
		end

	is_free_operator: BOOLEAN is
			-- Note: this feature which checks the correctness of free
			-- operator notation is not used by the parser but is defined just
			-- to make the assertion code clear and nice.
		local
			set: STRING; i: INTEGER
		do
			set := once "+-*/\=<>@#|&~"
			Result := set.has(to_string.first)
			from
				i := 2
			until
				not Result or else i > to_string.count - 1
			loop
				if set.has(to_string.item(i)) then
				else
					inspect
						to_string.item(i)
					when '.', '?', '{', '}' then
					else
						Result := False
					end
				end
				i := i + 1
			end
			if Result then
				Result := set.has(to_string.first)
			end
		end

	declaration_in (buffer: STRING) is
		require
			buffer /= Void
		do
			inspect
				name_type
			when C_simple_feature_name then
				buffer.append(to_string)
			when C_infix_name then
				buffer.append(fz_infix)
				buffer.extend(' ')
				buffer.extend('%"')
				buffer.append(to_string)
				buffer.extend('%"')
			when C_prefix_name then
				buffer.append(fz_prefix)
				buffer.extend(' ')
				buffer.extend('%"')
				buffer.append(to_string)
				buffer.extend('%"')
			when C_frozen_name then
				feature_name.declaration_in(buffer)
			end
		end

	declaration_pretty_print is
		do
			inspect
				name_type
			when C_simple_feature_name then
				pretty_printer.put_string(to_string)
			when C_infix_name then
				pretty_printer.keyword(fz_infix)
				pretty_printer.put_character('%"')
				pretty_printer.put_string(to_string)
				pretty_printer.put_character('%"')
			when C_prefix_name then
				pretty_printer.keyword(fz_prefix)
				pretty_printer.put_character('%"')
				pretty_printer.put_string(to_string)
				pretty_printer.put_character('%"')
			when C_frozen_name then
				pretty_printer.keyword(once "frozen")
				feature_name.declaration_pretty_print
			end
		end

	pretty (indent_level: INTEGER) is
		do
			inspect
				name_type
			when C_simple_feature_name, C_infix_name, C_prefix_name then
				pretty_printer.put_string(to_string)
			when C_frozen_name then
				feature_name.pretty(indent_level)
			end
		end

	pretty_target (indent_level: INTEGER) is
		do
			check
				C_simple_feature_name = name_type
			end
			pretty_printer.put_string(to_string)
			pretty_printer.put_character('.')
		end

	short (type: TYPE) is
		local
			i: INTEGER; c: CHARACTER
		do
			inspect
				name_type
			when C_simple_feature_name then
				short_printer.hook(once "Bsfn")
				from
					i := 1
				until
					i > to_string.count
				loop
					c := to_string.item(i)
					if c = '_' then
						short_printer.hook_or(once "Usfn", once "_")
					else
						short_printer.put_character(c)
					end
					i := i + 1
				end
				short_printer.hook(once "Asfn")
			when C_infix_name then
				short_printer.put_infix_name(once "Bifn", once "infix %"", once "Aifn", once "%"", Current)
			when C_prefix_name then
				short_printer.hook_or(once "Bpfn", once "prefix %"")
				from
					i := 1
				until
					i > to_string.count
				loop
					c := to_string.item(i)
					short_printer.put_character(c)
					i := i + 1
				end
				short_printer.hook_or(once "Apfn", once "%"")
			when C_frozen_name then
				feature_name.short(type)
			end
		end

	short_target (type: TYPE) is
		do
			check
				C_simple_feature_name = name_type
			end
			short(type)
			short_printer.put_dot
		end

	jvm_assign_creation (type: TYPE) is
		local
			space, idx: INTEGER; rf2: like run_feature_2; ca: like code_attribute
		do
			rf2 := run_feature_2
			ca := code_attribute
			space := rf2.result_type.jvm_stack_space
			ca.opcode_aload_0
			if space = 1 then
				ca.opcode_swap
			else
				ca.opcode_dup_x2
				ca.opcode_pop
			end
			idx := constant_pool.idx_fieldref(rf2)
			ca.opcode_putfield(idx, - (space + 1))
		end

	jvm_assign (type: TYPE) is
		local
			space, idx: INTEGER; rf2: like run_feature_2; ca: like code_attribute; wa: ARRAY[RUN_FEATURE_2]
			lt: LIVE_TYPE; flag: BOOLEAN
		do
			check
				C_simple_feature_name = name_type
			end
			rf2 := run_feature_2
			ca := code_attribute
			space := rf2.result_type.jvm_stack_space
			lt := rf2.result_type.live_type
			flag := lt.name.to_string.has_substring("NATIVE_ARRAY")
			if rf2.result_type.is_kernel_expanded or rf2.result_type.is_reference or flag then
				ca.opcode_aload_0
				if space = 1 then
					ca.opcode_swap
				else
					ca.opcode_dup_x2
					ca.opcode_pop
				end
				idx := constant_pool.idx_fieldref(rf2)
				ca.opcode_putfield(idx, - (space + 1))
			elseif rf2.result_type.is_bit then
				idx := constant_pool.idx_methodref3(fz_java_util_bitset, fz_a6, fz_a7)
				ca.opcode_invokevirtual(idx, 0)
				ca.opcode_aload_0
				idx := constant_pool.idx_fieldref(rf2)
				ca.opcode_getfield(idx, space - 1)
				ca.opcode_dup
				idx := constant_pool.idx_methodref3(fz_java_util_bitset, fz_dc, fz_b1)
				ca.opcode_invokevirtual(idx, 0)
				ca.opcode_aload_0
				idx := constant_pool.idx_fieldref(rf2)
				ca.opcode_getfield(idx, space - 1)
				ca.opcode_swap
				idx := constant_pool.idx_class2(fz_java_util_bitset)
				ca.opcode_checkcast(idx)
				idx := constant_pool.idx_methodref3(fz_java_util_bitset, fz_dd, fz_b1)
				ca.opcode_invokevirtual(idx, 0)
			else
				ca.opcode_aload_0
				idx := constant_pool.idx_fieldref(rf2)
				ca.opcode_getfield(idx, space - 1)
				ca.opcode_swap
				wa := lt.writable_attributes
				jvm.fields_by_fields_expanded_copy(wa)
				ca.opcode_pop2
			end
		end

	compile_target_to_jvm, compile_to_jvm (type: TYPE) is
		do
			error_handler.add_position(start_position)
			error_handler.append(fz_jvm_error)
			error_handler.print_as_fatal_error
		end

	jvm_branch_if_false (type: TYPE): INTEGER is
		do
			compile_to_jvm(type)
			Result := code_attribute.opcode_ifeq
		end

	jvm_branch_if_true (type: TYPE): INTEGER is
		do
			compile_to_jvm(type)
			Result := code_attribute.opcode_ifne
		end

	compile_to_jvm_into (type: TYPE; dest: TYPE_MARK): INTEGER is
		do
			Result := standard_compile_to_jvm_into(type, dest)
		end

	collect (type: TYPE): TYPE is
		do
			-- Register the feature as being alive:
			Result := smart_eiffel.collect(type, feature_stamp, False)
		end

	accept (visitor: FEATURE_NAME_VISITOR) is
		do
			visitor.visit_feature_name(Current)
		end

feature {FEATURE_NAME, LIVE_TYPE}
	set_run_feature_2 (rf2: RUN_FEATURE_2) is
		require
			rf2 /= Void
		do
			run_feature_2 := rf2
		ensure
			run_feature_2 = rf2
		end

feature {RUN_FEATURE_3}
	run_feature_2: RUN_FEATURE_2
			-- The corresponding one when runnable.

feature {CLASS_TEXT, LIVE_TYPE}
	unknown_position (n: STRING) is
		require
			n /= Void
		local
			unknown: POSITION
		do
			basic_create_1(C_simple_feature_name, n, unknown)
		ensure
			to_string.is_equal(n)
		end

feature {FEATURE_NAME}
	set_feature_stamp (fs: like feature_stamp) is
		do
			feature_stamp := fs
		end

	set_feature_stamp_and_type (fs: like feature_stamp; dt: like declaration_type) is
		do
			feature_stamp := fs
			declaration_type := dt
		ensure
			feature_stamp = fs
			declaration_type = dt
		end

feature {E_ROUTINE}
	feature_name: like Current
			-- Used only for `C_frozen_name' to refer the corresponding one.

feature {}
	name_type: INTEGER
			-- Dynamic binding is no longer used for feature name in order to avoid covariance problems in hashable DICTIONARY/SET.

	C_simple_feature_name: INTEGER is 1

	C_infix_name: INTEGER is 2

	C_prefix_name: INTEGER is 3

	C_frozen_name: INTEGER is 4

	ordinary_name (hn: like name; sp: like start_position) is
			-- To create a simple ordinary feature name.
		require
			hn /= Void
			not sp.is_unknown
		do
			basic_create_0(C_simple_feature_name, hn, hn, sp)
		ensure
			name = hn
			start_position = sp
		end

	simple_feature_name (n: STRING; sp: like start_position) is
		require
			n /= Void
			not sp.is_unknown
		do
			basic_create_1(C_simple_feature_name, n, sp)
		ensure
			to_string.is_equal(n)
			start_position = sp
		end

	prefix_name (hn: like name; sp: like start_position) is
		require
			hn /= Void
			not sp.is_unknown
		local
			kn: like key_name
		do
			kn := string_aliaser.for_prefix(hn.to_string)
			basic_create_0(C_prefix_name, hn, kn, sp)
		ensure
			name = hn
			start_position = sp
		end

	infix_name (hn: like name; sp: like start_position) is
		require
			hn /= Void
			not sp.is_unknown
		local
			kn: like key_name
		do
			kn := string_aliaser.for_infix(hn.to_string)
			basic_create_0(C_infix_name, hn, kn, sp)
		ensure
			name = hn
			start_position = sp
		end

	frozen_name (fn: like feature_name) is
		require
			fn /= Void
		do
			feature_name := fn
			basic_create_0(C_frozen_name, fn.name, fn.key_name, fn.start_position)
		ensure
			feature_name = fn
		end

	basic_create_1 (nt: like name_type; n: STRING; sp: like start_position) is
		require
			n /= Void
		local
			hs: HASHED_STRING
		do
			hs := string_aliaser.hashed_string(n)
			basic_create_0(nt, hs, hs, sp)
		end

	basic_create_0 (nt: INTEGER; hn: like name; kn: like key_name; sp: like start_position) is
			-- The really basic final one.
		require
			hn /= Void
			kn /= Void
		do
			name_type := nt
			debug
				-- To check `name_type':
				inspect
					nt
				when C_simple_feature_name then
				when C_infix_name then
				when C_prefix_name then
				when C_frozen_name then
				end
			end
			start_position := sp
			name := hn
			key_name := kn
			hash_code := kn.hash_code
		end

	complete_name_memory: STRING

invariant
	name /= Void

	key_name /= Void

	hash_code = key_name.hash_code

	name_type > 0

	string_aliaser.registered_one(to_string)

	string_aliaser.registered_one(to_key)

	eiffel_parser.case_insensitive implies to_string.is_equal(to_string.as_lower)

end -- class FEATURE_NAME
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
