-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class IMPLICIT_CAST
	--
	-- To memorize an implicit legal cast from one type to another.
	-- This invisible expression is used for all kinds of implicit
	-- conversions:
	--   from INTEGER to REAL
	--   from INTEGER to DOUBLE,
	--   from REAL to DOUBLE,
	--   from INTEGER to reference INTEGER,
	--   from INTEGER_8 to reference INTEGER_8,
	--   ...
	--   from FOO to reference FOO,
	--   ...

inherit
	EXPRESSION

creation {ASSIGNMENT_HANDLER, CALL_1, IMPLICIT_CAST}
	make

feature {ANY}
	resolved_memory: TYPE
			-- The final type of Current.

	expression: EXPRESSION
			-- The wrapped `expression'.

	expression_type: TYPE
			-- The original one at the creation of this IMPLICIT_CAST.

	start_position: POSITION is
		do
			Result := expression.start_position
		end

	is_manifest_string: BOOLEAN is
		do
			Result := expression.is_manifest_string
		end

	extra_bracket_flag: BOOLEAN is
		do
			Result := expression.extra_bracket_flag
		end

	resolve_in (type: TYPE): TYPE is
		do
			Result := resolved_memory
		end

	is_void: BOOLEAN is
		do
			Result := expression.is_void
		end

	declaration_type: TYPE is
		do
			-- *** Result := expression.declaration_type
			Result := resolved_memory
		end

	is_writable: BOOLEAN is
		do
			Result := expression.is_writable
		end

	is_result: BOOLEAN is
		do
			Result := expression.is_result
		end

	is_implicit_current, is_current: BOOLEAN is
		do
			-- (Even if expression is a `Current', we must not forget this IMPLICIT_CAST.)
		end

	is_static: BOOLEAN is
		do
			Result := expression.is_static
		end

	compile_to_c (type: TYPE) is
		local
			i: INTEGER; conversion: STRING; true_conversion_flag: BOOLEAN
		do
			if expression_type.is_expanded then
				true_conversion_flag := resolved_memory.is_reference
			else
				true_conversion_flag := resolved_memory.is_expanded
			end
			if true_conversion_flag then
				conversion := assignment_handler.collect_actual_conversion(expression_type, resolved_memory)
			else
				conversion := once "/*IC*/"
				resolved_memory.canonical_type_mark.mapping_cast
			end
			cpp.put_string(conversion)
			cpp.put_character('(')
			if expression_type.is_reference then
				-- To promote automatically reference INTEGER_* type:
				cpp.put_character('(')
				from
					i := 1
				until
					conversion.item(i) = 't'
				loop
					cpp.put_character(conversion.item(i))
					i := i + 1
				end
				cpp.put_string(once "*)(")
				expression.compile_to_c(type)
				cpp.put_character(')')
			else
				expression.mapping_c_target(type, expression_type.canonical_type_mark)
			end
			cpp.put_character(')')
		end

	mapping_c_target (type: TYPE; target_type: TYPE_MARK) is
		do
			compile_to_c(type)
		end

	mapping_c_arg (type: TYPE; formal_arg_type: TYPE_MARK) is
		do
			compile_to_c(type)
		end

	safety_check is
		do
			expression.safety_check
		end

	compile_to_jvm (type: TYPE) is
		local
			space: INTEGER; expression_type_mark: TYPE_MARK
		do
			expression.compile_to_jvm(type)
			expression_type_mark := expression_type.canonical_type_mark
			space := expression_type_mark.jvm_convert_to(resolved_memory.canonical_type_mark)
		end

	compile_target_to_jvm (type: TYPE) is
		local
			space: INTEGER; expression_type_mark: TYPE_MARK
		do
			expression.compile_target_to_jvm(type)
			expression_type_mark := expression_type.canonical_type_mark
			space := expression_type_mark.jvm_convert_to(resolved_memory.canonical_type_mark)
		end

	compile_to_jvm_into (type: TYPE; dest: TYPE_MARK): INTEGER is
		do
			compile_to_jvm(type)
			Result := resolved_memory.canonical_type_mark.jvm_convert_to(dest)
		end

	jvm_branch_if_false (type: TYPE): INTEGER is
		do
			Result := expression.jvm_branch_if_false(type)
		end

	jvm_branch_if_true (type: TYPE): INTEGER is
		do
			Result := expression.jvm_branch_if_true(type)
		end

	use_current: BOOLEAN is
		do
			Result := expression.use_current
		end

	side_effect_free (type: TYPE): BOOLEAN is
		do
			Result := expression.side_effect_free(type)
		end

	specialize_in (new_type: TYPE): like Current is
		do
			check
				False -- (IMPLICIT_CAST are not written.)
			end
		end

	specialize_thru (parent_type: TYPE; inherit_item: INHERIT_TEXT_ITEM; new_type: TYPE): like Current is
		local
			e: like expression; implicit_cast: IMPLICIT_CAST
		do
			-- To avoid nested IMPLICIT_CAST:
			implicit_cast ?= expression
			if implicit_cast = Void then
				e := expression.specialize_thru(parent_type, inherit_item, new_type)
			else
				e := implicit_cast.expression.specialize_thru(parent_type, inherit_item, new_type)
			end
			if e = expression then
				Result := Current
			else
				Result := twin
				Result.set_expression(e)
			end
		end

	specialize_2 (type: TYPE): EXPRESSION is
		local
			e: like expression; type_reference: TYPE_REFERENCE
		do
			e := expression.specialize_2(type)
			type_reference ?= resolved_memory.canonical_type_mark
			if type_reference /= Void then
				-- Remove this previously added extra conversion to consider again the original `expression':
				Result := e
			elseif type.is_integer or else type.is_reference_integer then
				-- Because INTEGER_GENERAL is a very special case, we have to unwrap the IMPLICIT_CAST in order to give
				-- another chance for expression to match with INTEGER_8 / 16 / 32 / 64:
				Result := e
			elseif e = expression then
				Result := Current
			else
				create {IMPLICIT_CAST} Result.make(e, e.resolve_in(type), resolved_memory)
			end
		end

	has_been_specialized: BOOLEAN is
		do
			Result := expression.has_been_specialized
		end

	collect (type: TYPE): TYPE is
		do
			Result := expression.collect(type)
			assignment_handler.collect(expression.start_position, Result, resolved_memory)
			Result := resolved_memory
		end

	adapt_for (type: TYPE): like Current is
		local
			e: like expression; call: CALL; target_type: TYPE
		do
			e := expression.adapt_for(type)
			call ?= e
			if call /= Void then
				target_type := call.extra_local_expanded(type)
				if target_type /= Void then
					smart_eiffel.register_extra_local_expanded(target_type)
				end
			end
			if e = expression then
				Result := Current
			else
				Result := twin
				Result.set_expression(e)
			end
		end

	non_void_no_dispatch_type (type: TYPE): TYPE is
		do
			-- *** To be done. *** (Dom. june 18th 2004) ***
		end

	simplify (type: TYPE): EXPRESSION is
		local
			e: like expression; ic: INTEGER_CONSTANT; rc: REAL_CONSTANT; et: like expression_type
		do
			-- Going back to the INTEGER_CONSTANT canonical form:
			ic ?= expression
			if ic /= Void and then resolved_memory.is_integer then
				smart_eiffel.magic_count_increment
				ic := ic.twin
				ic.set_result_type(resolved_memory.canonical_type_mark)
				Result := ic
			end
			-- Going back to the REAL_CONSTANT canonical form:
			if Result = Void then
				rc ?= expression
				if rc /= Void and then (resolved_memory.is_real or else resolved_memory.is_double) then
					smart_eiffel.magic_count_increment
					rc := rc.twin
					rc.set_result_type(resolved_memory.canonical_type_mark)
					Result := rc
				end
			end
			-- Finally, the general scheme:
			if Result = Void then
				e := expression.simplify(type)
				et := e.resolve_in(type)
				if e = expression and then et = expression_type then
					Result := Current
				else
					smart_eiffel.magic_count_increment
					create {IMPLICIT_CAST} Result.make(e, et, resolved_memory)
				end
			end
		end

	accept (visitor: IMPLICIT_CAST_VISITOR) is
		do
			visitor.visit_implicit_cast(Current)
		end

	pretty (indent_level: INTEGER) is
		do
			expression.pretty(indent_level)
		end

	bracketed_pretty (indent_level: INTEGER) is
		do
			expression.bracketed_pretty(indent_level)
		end

	pretty_target (indent_level: INTEGER) is
		do
			expression.pretty_target(indent_level)
		end

	short (type: TYPE) is
		do
			expression.short(type)
		end

	short_target (type: TYPE) is
		do
			expression.short_target(type)
		end

	precedence: INTEGER is
		do
			Result := expression.precedence
		end

	jvm_assign_creation (type: TYPE) is
		do
			expression.jvm_assign_creation(type)
		end

	jvm_assign (type: TYPE) is
		do
			expression.jvm_assign(type)
		end

feature {IMPLICIT_CAST}
	set_expression (e: like expression) is
		do
			expression := e
		end

feature {}
	make (e: like expression; et: like expression_type; rm: like resolved_memory) is
		require
			no_implicit_cast_of_implicit_cast: not e.generating_type.is_equal(once "IMPLICIT_CAST")
		local
			tr1, tr2: TYPE_REFERENCE
		do
			debug
				if et.is_reference and then rm.is_reference then
					tr1 ?= et.canonical_type_mark
					tr2 ?= rm.canonical_type_mark
					check
					-- Is it a really strange reference to reference conversion ?
						tr1 /= Void
						tr2 /= Void
					end
				end
			end
			resolved_memory := rm
			expression := e
			expression_type := et
		ensure
			expression /= Void
			expression_type = et
			resolved_memory = rm
		end

invariant
	resolved_memory /= Void

	expression /= Void

	expression_type /= Void

	no_implicit_cast_of_implicit_cast: not expression.generating_type.is_equal(once "IMPLICIT_CAST")

end -- class IMPLICIT_CAST
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
