-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class OPEN_OPERAND
	--
	-- An open operand inside some AGENT_CREATION expression. Most often, an open operand is a single question mark
	-- inside some agent creation expression like in:  "agent foo(?)"
	-- But, keep in mind that an open operand can also be some type mark inside curly brackets like in:
	--                                                 "agent foo({INTEGER})"
	-- Note: the case of an open target is not handled with an OPEN_OPERAND object, but with a FAKE_TARGET object
	-- (check in the `eiffel_parser' code).
	--

inherit
	EXPRESSION

creation {EIFFEL_PARSER, FORMAL_ARG_LIST}
	question_mark

creation {EIFFEL_PARSER}
	type_holder

feature {ANY}
	start_position: POSITION
			-- Of the `?' or the '{' character.

	curly_type: TYPE_MARK
			-- If any, provided by the `eiffel_parser'. The one written between {} when the ? notation is
			-- not used.

	rank: INTEGER
			-- The `rank' of the corresponding open argument in the actual argument list .

	formal_type: TYPE
			-- The formal type in the corresponding feature.

	is_current: BOOLEAN is False

	is_implicit_current: BOOLEAN is False

	is_manifest_string: BOOLEAN is False

	is_void: BOOLEAN is False

	is_result: BOOLEAN is False

	is_writable: BOOLEAN is False

	once_pre_computable: BOOLEAN is False

	is_static: BOOLEAN is False

	extra_bracket_flag: BOOLEAN is False

	specialize_in (type: TYPE): like Current is
		local
			ct: like curly_type
		do
			if curly_type = Void then
				Result := Current
			else
				ct := curly_type.specialize_in(type)
				Result := current_or_twin_init(ct, formal_type)
			end
		end

	specialize_thru (parent_type: TYPE; inherit_item: INHERIT_TEXT_ITEM; new_type: TYPE): like Current is
		local
			ct: like curly_type
		do
			if curly_type = Void then
				Result := Current
			else
				ct := curly_type.specialize_thru(parent_type, inherit_item, new_type)
				Result := current_or_twin_init(ct, formal_type)
			end
		end

	specialize_2 (type: TYPE): like Current is
		local
			curly: TYPE; unknow_position: POSITION
		do
			if formal_type = Void then
				error_handler.add_position(start_position)
				error_handler.append("Open operand out of agent scope. (Missing %"agent%" keyword?)")
				error_handler.print_as_fatal_error
			end
			if curly_type /= Void then
				curly := curly_type.resolve_in(type)
				if not curly.is_a(formal_type, start_position, unknow_position) then
					error_handler.append(" (See also the next error message.)")
					error_handler.print_as_fatal_error
					error_handler.add_position(start_position)
					error_handler.append("Open argument constraint not satisfied.")
					error_handler.print_as_fatal_error
				end
			end
			Result := Current
		end

	declaration_type: TYPE is
		do
			Result := formal_type
		end

	collect (type: TYPE): TYPE is
		local
			curly: TYPE
		do
			if curly_type /= Void then
				curly := curly_type.resolve_in(type)
				assignment_handler.collect(curly_type.start_position, curly, formal_type)
			end
			Result := resolve_in(type)
		end

	side_effect_free (type: TYPE): BOOLEAN is
		do
			Result := True
		end

	resolve_in (type: TYPE): TYPE is
		do
			Result := formal_type
		end

	has_been_specialized: BOOLEAN is
		do
			if curly_type = Void then
				Result := True
			else
				Result := curly_type.has_been_specialized
			end
		end

	adapt_for (t: TYPE): like Current is
		do
			Result := Current
		end

	bracketed_pretty, pretty (indent_level: INTEGER) is
		do
			if curly_type = Void then
				pretty_printer.put_character('?')
			else
				pretty_printer.put_character('{')
				pretty_printer.put_type_mark(curly_type)
				pretty_printer.put_character('}')
			end
		end

	pretty_target (indent_level: INTEGER) is
		do
			check
				False
			end
		end

	short (type: TYPE) is
		do
			if curly_type /= Void then
				short_printer.hook_or("open_curly_bracket", once "{")
				curly_type.short(type)
				short_printer.hook_or("close_curly_bracket", once "}")
			else
				short_printer.hook_or("question_mark", once "?")
			end
		end

	short_target (type: TYPE) is
		do
			check
				False
			end
			-- An OPEN_OPERAND is never a target !
		end

	precedence: INTEGER is
		do
			Result := atomic_precedence
		end

	non_void_no_dispatch_type (type: TYPE): TYPE is
		do
			-- *** To be done. *** (Dom. june 18th 2004) ***
		end

	simplify (type: TYPE): OPEN_OPERAND is
		do
			Result := Current
		end

	safety_check is
		do
		end

	compile_to_c (type: TYPE) is
		do
			cpp.put_character('a')
			cpp.put_integer(rank)
		end

	mapping_c_target (type: TYPE; ft: TYPE_MARK) is
		do
			check
				False
			end
		end

	mapping_c_arg (type: TYPE; formal_arg_type: TYPE_MARK) is
		do
			compile_to_c(type)
		end

	use_current: BOOLEAN is False

	jvm_assign_creation, jvm_assign (type: TYPE) is
		do
			check
				False
			end
		end

	compile_to_jvm_into (type: TYPE; dest: TYPE_MARK): INTEGER is
		do
			not_yet_implemented
		end

	compile_target_to_jvm, compile_to_jvm (type: TYPE) is
		do
			not_yet_implemented
		end

	jvm_branch_if_false (type: TYPE): INTEGER is
		do
			not_yet_implemented
		end

	jvm_branch_if_true (type: TYPE): INTEGER is
		do
			not_yet_implemented
		end

	accept (visitor: OPEN_OPERAND_VISITOR) is
		do
			visitor.visit_open_operand(Current)
		end

feature {OPEN_OPERAND}
	init (ct: like curly_type; af: like formal_type) is
		do
			curly_type := ct
			formal_type := af
		ensure
			curly_type = ct
			formal_type = af
		end

feature {AGENT_CREATION, EFFECTIVE_ARG_LIST}
	notify (target_type: TYPE; af: ANONYMOUS_FEATURE; r: like rank): like Current is
			-- Assign `Current' with some `formal_type' and `rank'.
		require
			af /= Void
			r >= 1
		local
			ft: like formal_type
		do
			ft := af.arguments.type_mark(r).resolve_in(target_type)
			if formal_type = Void then
				formal_type := ft
				check
					rank = 0
				end
				rank := r
				Result := Current
			else
				check
					rank = r
				end
				Result := current_or_twin_init(curly_type, ft)
			end
		ensure
			Result.formal_type /= Void
			Result.rank = r
		end

feature {}
	current_or_twin_init (ct: like curly_type; ft: like formal_type): like Current is
		do
			if ct = curly_type and then ft = formal_type then
				Result := Current
			else
				Result := twin
				Result.init(ct, ft)
			end
		ensure
			Result.curly_type = ct
			Result.formal_type = ft
		end

	question_mark (sp: like start_position) is
			-- To create the classic ? open operand.
		require
			not sp.is_unknown
		do
			start_position := sp
		ensure
			start_position = sp
		end

	type_holder (sp: like start_position; ct: like curly_type) is
			-- To create the curly bracketed type-holder form.
		require
			not sp.is_unknown
			ct /= Void
		do
			start_position := sp
			curly_type := ct
		ensure
			start_position = sp
			curly_type = ct
		end

invariant
	formal_type /= Void implies rank > 0

end -- class OPEN_OPERAND
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
