-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class REAL_CONSTANT
	--
	-- For Manifest Constant REAL as well as DOUBLE.
	--

inherit
	BASE_TYPE_CONSTANT

creation {ANY}
	make, from_double

feature {ANY}
	is_static: BOOLEAN is True

	to_string: STRING
			-- Cleanly written view of the constant (ANSI C compatible).

	value: DOUBLE is
		do
			Result := to_string.to_double
		end

	compile_to_c (type: TYPE) is
		do
			cpp.put_string(to_string)
		end

	compile_target_to_jvm, compile_to_jvm (type: TYPE) is
		do
			if result_type.is_double then
				code_attribute.opcode_push_as_double(to_string)
			else
				code_attribute.opcode_push_as_float(to_string)
			end
		end

	jvm_branch_if_false, jvm_branch_if_true (type: TYPE): INTEGER is
		do
			check
				False
			end
		end

	compile_to_jvm_into (type: TYPE; dest: TYPE_MARK): INTEGER is
		do
			if dest.is_real then
				code_attribute.opcode_push_as_float(to_string)
				Result := 1
			elseif dest.is_double then
				code_attribute.opcode_push_as_double(to_string)
				Result := 2
			else
				Result := standard_compile_to_jvm_into(type, dest)
			end
		end

	simplify_1_, simplify_2: like Current is
		do
			Result := Current
		end

feature {IMPLICIT_CAST, TMP_FEATURE}
	set_result_type (rt: TYPE_MARK) is
		require
			rt.is_double or else rt.is_real
		do
			result_type := rt
		ensure
			result_type = rt
		end

feature {ANY}
	accept (visitor: REAL_CONSTANT_VISITOR) is
		do
			visitor.visit_real_constant(Current)
		end

feature {}
	make (sp: like start_position; ts: like to_string) is
		require
			not sp.is_unknown
			ts.is_double
			not string_aliaser.registered_one(ts)
		local
			pos, exponent: INTEGER
			stop: BOOLEAN
		do
			start_position := sp
			to_string := ts
			if smart_eiffel.pretty_flag then
			elseif smart_eiffel.short_or_class_check_flag then
			else
				-- Kind of normalisation to allow easy comparison.
				-- *** This seems to be used in CODE_ATTRIBUTE and not 
				-- *** much else. Maybe we should get rid of the whole 
				-- *** normalisation. <FM-14/08/2004>
				-- Part 1: Remove the exponent if there is one.
				pos := to_string.first_index_of('e')
				if pos = 0 then
					pos := to_string.first_index_of('E')
				end
				if pos /= 0 then
					exponent := to_string.substring(pos + 1, to_string.count).to_integer
					to_string.keep_head(pos-1)
				end
				-- Part 2: find out if there is a decimal point. Remove it
				-- if weed need to move it because of the exponent.
				pos := to_string.first_index_of('.')
				if exponent /= 0 and then pos /= 0 then
					exponent := exponent - (to_string.count - pos)
					to_string.remove(pos)
					pos := 0
				end
				-- Part 3: normalize the decimal part of the number.
				if pos /= 0 then
					-- If there is a decimal point, just get rid of 
					-- trailing zeroes.
					check
						exponent = 0
					end
					from
					until
						to_string.last /= '0'
					loop
						to_string.remove_last(1)
					end
					if to_string.last = '.' then
						to_string.extend('0')
					end
				else
					-- If there is no decimal point, add one. Decide what 
					-- to do with the trailing zeroes according to the exponent.
					if exponent >= 0 then
						to_string.extend_multiple('0', exponent)
						to_string.append(once ".0")
					else
						from
							stop := False
						until
							stop
						loop
							if to_string.last = '0' then
								if to_string.count > 1 then
									to_string.remove_last(1)
									exponent := exponent + 1
								else
									exponent := 0
								end
								if exponent = 0 then
									to_string.append(once ".0")
									stop := True
								end
							else
								pos := to_string.count + exponent + 1
								if pos <= 1 then
									to_string.precede_multiple('0', 3 - pos)
									pos := 2
									to_string.put('.', pos)
								else
									to_string.insert_character('.', pos)
								end
								stop := True
							end
						end
					end
				end
				-- Part 4: normalize the integral part of the number.
				if pos = 1 then
					check
						to_string.first_index_of('.') = 1
					end
					to_string.precede('0')
				else
					check
						to_string.first_index_of('.') /= 1
					end
					from
						pos := 1
					until
						to_string.item(pos) /= '0'
					loop
						pos := pos + 1
					end
					if to_string.item(pos) = '.' then
						pos := pos - 1
					end
					if pos > 1 then
						to_string.remove_head(pos - 1)
					end
				end
			end
			-- Setting the `result_type':
			create {TYPE_REAL} result_type.make(start_position)
		ensure
			start_position = sp
			to_string = ts
		end

	from_double (sp: like start_position; double: DOUBLE; rt: like result_type) is
		require
			rt.is_double or else rt.is_real
		do
			make(sp, double.to_string_format(10))
			--                               **
			-- We need to redesign REAL/DOUBLE notation... :-(
			result_type := rt
		end

end -- class REAL_CONSTANT
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
