-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class ASSIGNMENT
	--
	-- For instruction like :
	--                          foo := bar
	--                          foo := bar + 1
	--
	--

inherit
	INSTRUCTION

creation {EIFFEL_PARSER, ASSIGNMENT, ASSIGNMENT_ATTEMPT, PROC_CALL_1}
	make

creation {E_PROCEDURE, ASSIGNMENT}
	attribute_writer
	
feature {ANY}
	left_side: EXPRESSION

	right_side: EXPRESSION

	end_mark_comment: BOOLEAN is False

	side_effect_free (type: TYPE): BOOLEAN is
		do
		end

	simplify (type: TYPE): INSTRUCTION is
		local
			rs: like right_side
		do
			rs := right_side.simplify(type)
			if rs /= right_side then
				smart_eiffel.magic_count_increment
				create {ASSIGNMENT} Result.attribute_writer(left_side, rs)
			else
				Result := Current
			end
		end

	compile_to_c (type: TYPE) is
		local
			cast_t0: BOOLEAN; right_type: TYPE
		do
			right_type := right_side.resolve_in(type)
			if ace.no_check then
				cpp.put_trace_or_sedb_instruction(start_position)
			end
			if right_side.is_current then
				if left_side.resolve_in(type).is_reference then
					cast_t0 := right_type.is_reference
				end
			end
			left_side.compile_to_c(type)
			cpp.put_character('=')
			if cast_t0 then
				cpp.put_string(once "((T0*)(")
			elseif right_type.is_boolean then
				cpp.put_string(once "((T6)(")
			end
			right_side.compile_to_c(type)
			if cast_t0 or else right_type.is_boolean then
				cpp.put_string(once "))")
			end
			cpp.put_string(fz_00)
		end

	compile_to_jvm (type: TYPE) is
		do
			right_side.compile_to_jvm(type)
			left_side.jvm_assign(type)
		end

	use_current: BOOLEAN is
		do
			Result := left_side.use_current
			Result := Result or else right_side.use_current
		end

	start_position: POSITION is
		do
			Result := left_side.start_position
		end

	specialize_in (type: TYPE): like Current is
		local
			l, r: EXPRESSION
		do
			l := left_side.specialize_in(type)
			r := right_side.specialize_in(type)
			if r = right_side and then l = left_side then
				Result := Current
			else
				create Result.make(l, r)
			end
		end

	specialize_thru (parent_type: TYPE; inherit_item: INHERIT_TEXT_ITEM; new_type: TYPE): like Current is
		local
			l, r: EXPRESSION
		do
			l := left_side.specialize_thru(parent_type, inherit_item, new_type)
			r := right_side.specialize_thru(parent_type, inherit_item, new_type)
			if r = right_side and then l = left_side then
				Result := Current
			else
				create Result.make(l, r)
			end
		end

	specialize_2 (type: TYPE): like Current is
		local
			l, r, rhs: EXPRESSION; lt, rt: TYPE
		do
			l := left_side.specialize_2(type)
			r := right_side.specialize_2(type)
			lt := l.resolve_in(type)
			rt := r.resolve_in(type)
			-- Theoretically, validity checking should be done only once using the sole `declaration_type'.
			-- In practice, using the `declaration_type' needs more computation ... and this solution appears
			-- to catch errors which are not catched using the `declaration_type'. (Dom. july 12th 2004)
			if r.is_void and then lt.is_expanded then
				fatal_error_for("Void cannot be assigned to an expanded entity. ", type, lt, rt)
			end
			if lt.is_expanded and then rt.is_reference then
				fatal_error_for("Assignment of a reference type into an expanded type is not allowed. ", type, lt, rt)
			end
			if not rt.is_a(lt, l.start_position, r.start_position) then
				fatal_error_for(" Bad assignment (VJAR). ", type, lt, rt)
			end
			if rt.is_reference then
				separateness_consistency_rule_3_check(type, lt, r)
			end
			if rt = lt and then r = right_side and then l = left_side then
				Result := Current
			else
				if rt = lt then
					rhs := r
				else
					rhs := assignment_handler.implicit_cast(r, rt, lt)
				end
				create Result.make(l, rhs)
			end
		end

	has_been_specialized: BOOLEAN is
		do
			Result := left_side.has_been_specialized
			Result := Result and right_side.has_been_specialized
		end

	safety_check is
		do
			right_side.safety_check
		end

	pretty (indent_level: INTEGER) is
		local
			semi_colon_flag: BOOLEAN
		do
			pretty_printer.set_indent_level(indent_level)
			semi_colon_flag := pretty_printer.semi_colon_flag
			left_side.pretty(indent_level)
			pretty_printer.put_string(once " := ")
			pretty_printer.set_semi_colon_flag(False)
			right_side.pretty(indent_level)
			pretty_printer.set_semi_colon_flag(semi_colon_flag)
			if semi_colon_flag then
				pretty_printer.put_character(';')
			end
			pretty_printer.set_indent_level(0)
		end

	accept (visitor: ASSIGNMENT_VISITOR) is
		do
			visitor.visit_assignment(Current)
		end

	verify_scoop (type: TYPE; allowed: FORMAL_ARG_LIST) is
		local
			dummy: BOOLEAN
		do
			dummy := left_side.verify_scoop(type, allowed)
			dummy := right_side.verify_scoop(type, allowed)
		end

	collect (t: TYPE) is
		local
			l, r: TYPE
		do
			l := left_side.collect(t)
			r := right_side.collect(t)
			assignment_handler.collect(right_side.start_position, r, l)
		end

	adapt_for (t: TYPE): like Current is
		local
			l: like left_side; r: like right_side
		do
			l := left_side.adapt_for(t)
			r := right_side.adapt_for(t)
			if l = left_side and then r = right_side then
				Result := Current
			else
				create Result.make(l, r)
			end
		end

feature {}
	separateness_consistency_rule_3_check (type, lt: TYPE; r: EXPRESSION) is
		local
			c: CALL; left_type: TYPE
		do
			c ?= r
			if c /= Void then
				if c.target.resolve_in(type).is_separate and then not c.target.is_current and then not lt.is_separate then
					error_handler.add_position(left_side.start_position)
					error_handler.append("Separateness Consistency Rule violated%N(OOSC2 p. 1025): invalid assignment. (See the %
                                    %2 following messages for more details.)")
					error_handler.print_as_error
					error_handler.add_position(c.start_position)
					error_handler.append("This call is separate (because the target is separate).")
					error_handler.print_as_error
					left_type := left_side.resolve_in(type)
					error_handler.add_position(left_type.canonical_type_mark.start_position)
					-- *** Using the `start_position'of a canonical_type_mark is probably a bad idea.
					-- *** (Dom. july 13th 2004) ***
					error_handler.append("In this situation, the type of the left-hand-side must be separate too.")
					error_handler.print_as_fatal_error
				end
			end
		end

	fatal_error_for (msg: STRING; context_type, lt, rt: TYPE) is
		do
			error_handler.append(msg)
			error_handler.add_position(right_side.start_position)
			error_handler.add_position(left_side.start_position)
			if left_side.start_position.class_text /= context_type.class_text then
				error_handler.append("Error detected while checking this code in the ")
				error_handler.append(context_type.name.to_string)
				error_handler.append(" context. ")
			end
			error_handler.append("Cannot assign ")
			error_handler.add_expression(right_side)
			error_handler.append(" which is of type ")
			error_handler.append(rt.name.to_string)
			error_handler.append(" into ")
			error_handler.add_expression(left_side)
			error_handler.append(" which is of type ")
			error_handler.append(lt.name.to_string)
			error_handler.append(".")
			error_handler.print_as_fatal_error
		end

	make (ls: like left_side; rs: like right_side) is
			-- Note: this creation procedure is for example called by the `eiffel_parser' which is in charge 
			-- of checking that `ls' is actually a writable entity. (See also `attribute_writer'.)
		require
			ls.is_writable
			not ls.start_position.is_unknown
			rs /= Void
		do
			left_side := ls
			right_side := rs
		ensure
			left_side = ls
			right_side = rs
		end

	attribute_writer (ls: like left_side; rs: like right_side) is
			-- Note: this creation procedure is called to inline a procedure which is actually an attribute 
			-- setter, hence the `is_writable' property is not necessary. (See also `make'.)
		require
			ls /= Void
			rs /= Void
		do
			left_side := ls
			right_side := rs
		ensure
			left_side = ls
			right_side = rs
		end
	
invariant
	left_side /= Void

	right_side /= Void

end -- class ASSIGNMENT
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
