-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class PRECOMPUTABLE_ROUTINE_DETECTOR
	--
	-- The goal of this VISITOR is to determine which kind of code we can run
	-- during the precomputation of some once routine.
	--

inherit
	VISITOR
	CODE_VISITOR
	ANONYMOUS_FEATURE_VISITOR
	NATIVE_VISITOR
	LOCAL_VAR_LIST_VISITOR
	EFFECTIVE_ARG_LIST_VISITOR
insert
	GLOBALS

creation {ONCE_ROUTINE_POOL}
	make

feature {ANY}
	visit_e_procedure (visited: E_PROCEDURE) is
		do
			if visited.local_vars /= Void then
				visited.local_vars.accept(Current)
			end
			if precomputable and then visited.routine_body /= Void then
				visited.routine_body.accept(Current)
			end
			if precomputable and then visited.rescue_compound /= Void then
				precomputable := False
			end
		end

	visit_e_function (visited: E_FUNCTION) is
		do
			if visited.local_vars /= Void then
				visited.local_vars.accept(Current)
			end
			if precomputable and then visited.routine_body /= Void then
				visited.routine_body.accept(Current)
			end
			if precomputable and then visited.rescue_compound /= Void then
				precomputable := False
			end
		end

	visit_external_procedure (visited: EXTERNAL_PROCEDURE) is
		do
			visited.native.accept(Current)
		end

	visit_external_function (visited: EXTERNAL_FUNCTION) is
		do
			visited.native.accept(Current)
		end

	visit_deferred_procedure (visited: DEFERRED_PROCEDURE) is
		do
			precomputable := False
		end

	visit_deferred_function (visited: DEFERRED_FUNCTION) is
		do
			precomputable := False
		end

	visit_once_procedure (visited: ONCE_PROCEDURE) is
		do
			precomputable := False
		end

	visit_once_function (visited: ONCE_FUNCTION) is
		do
			precomputable := False
		end

	visit_native_built_in (visited: NATIVE_BUILT_IN) is
		local
			n: STRING
		do
			n := anonymous_feature_stack.last.first_name.to_string
			if n = as_and_then then
			elseif n = as_calloc then
			elseif n = as_copy then
			elseif n = as_code then
			elseif n = as_ge then
			elseif n = as_gt then
			elseif n = as_implies then
			elseif n = as_is_equal then
			elseif n = as_is_not_null then
			elseif n = as_item then
			elseif n = as_lt then
			elseif n = as_le then
			elseif n = as_sharp_minus then
			elseif n = as_sharp_muls then
			elseif n = as_object_size then
			elseif n = as_or_else then
			elseif n = as_put then
			elseif n = as_sharp_plus then
			elseif n = as_standard_copy then
			elseif n = as_standard_is_equal then
			elseif n = as_standard_twin then
			elseif n = as_to_integer then
			elseif n = as_to_pointer then
			elseif n = as_twin then
			else
				-- I don't think that other external "built_in" are dangerous, but I do not want to take
				-- risks right now. Also note that i/o are implemented with external "plug_in".
				precomputable := False
			end
		end

	visit_native_plug_in (visited: NATIVE_PLUG_IN) is
		do
			-- We do not trust outside component:
			precomputable := False
		end

	visit_native_java (visited: NATIVE_JAVA) is
		do
			-- (We do not trust Java at all :-)
			precomputable := False
		end

	visit_native_c_plus_plus (visited: NATIVE_C_PLUS_PLUS) is
		do
			-- (We do not trust C++ more :-)
			precomputable := False
		end

	visit_native_c (visited: NATIVE_C) is
		do
			precomputable := False
		end

	visit_cst_att_real (visited: CST_ATT_REAL) is
		do
		end

	visit_cst_att_boolean (visited: CST_ATT_BOOLEAN) is
		do
		end

	visit_cst_att_double (visited: CST_ATT_DOUBLE) is
		do
		end

	visit_cst_att_integer (visited: CST_ATT_INTEGER) is
		do
		end

	visit_cst_att_unique (visited: CST_ATT_UNIQUE) is
		do
		end

	visit_cst_att_string (visited: CST_ATT_STRING) is
		do
		end

	visit_cst_att_character (visited: CST_ATT_CHARACTER) is
		do
		end

	visit_cst_att_bit (visited: CST_ATT_BIT) is
		do
		end

	visit_local_var_list (visited: LOCAL_VAR_LIST) is
		local
			i: INTEGER; type_mark: TYPE_MARK; type: TYPE
		do
			from
				i := visited.count
			until
				not precomputable or else i = 0
			loop
				type_mark := visited.type_mark(i)
				type := type_mark.resolve_in(current_dynamic_type)
				if type.is_user_expanded then
					visit_anonymous_feature(type.live_type.default_create_stamp, type)
				end
				i := i - 1
			end
		end

	visit_effective_arg_list (visited: EFFECTIVE_ARG_LIST) is
		local
			i: INTEGER
		do
			from
				i := visited.count
			until
				not precomputable or else i = 0
			loop
				visited.expression(i).accept(Current)
				i := i - 1
			end
		end

	visit_comment (visited: COMMENT) is
		do
		end

	visit_fake_tuple (visited: FAKE_TUPLE) is
		do
			precomputable := False
		end

	visit_manifest_array (visited: MANIFEST_ARRAY) is
		local
			i: INTEGER; list: FAST_ARRAY[EXPRESSION]
		do
			from
				list := visited.list
				i := list.upper
			until
				not precomputable or else i < list.lower
			loop
				list.item(i).accept(Current)
				i := i - 1
			end
		end

	visit_manifest_tuple (visited: MANIFEST_TUPLE) is
		do
			precomputable := False
		end

	visit_e_debug (visited: E_DEBUG) is
		do
			precomputable := False
		end

	visit_e_check (visited: E_CHECK) is
		do
			precomputable := False
		end

	visit_precursor_instruction (visited: PRECURSOR_INSTRUCTION) is
		do
			precomputable := False
		end

	visit_precursor_expression (visited: PRECURSOR_EXPRESSION) is
		do
			precomputable := False
		end

	visit_e_retry (visited: E_RETRY) is
		do
			precomputable := False
		end

	visit_compound (visited: COMPOUND) is
		local
			i: INTEGER; list: FAST_ARRAY[INSTRUCTION]
		do
			from
				list := visited.list
				i := list.upper
			until
				not precomputable or else i < list.lower
			loop
				list.item(i).accept(Current)
				i := i - 1
			end
		end

	visit_call (target: EXPRESSION; feature_stamp: FEATURE_STAMP) is
		local
			type: TYPE
		do
			if target.is_current then
				visit_anonymous_feature(feature_stamp, current_dynamic_type)
			else
				type := target.non_void_no_dispatch_type(current_dynamic_type)
				if type = Void then
					precomputable := False
				else
					visit_anonymous_feature(feature_stamp, type)
				end
			end
		end

	visit_proc_call_0 (visited: PROC_CALL_0) is
		do
			visit_call(visited.target, visited.feature_stamp)
			if precomputable then
				visited.target.accept(Current)
			end
		end

	visit_proc_call_1 (visited: PROC_CALL_1) is
		do
			visit_call(visited.target, visited.feature_stamp)
			if precomputable then
				visited.target.accept(Current)
			end
			if precomputable then
				visited.arg1.accept(Current)
			end
		end

	visit_proc_call_n (visited: PROC_CALL_N) is
		do
			visit_call(visited.target, visited.feature_stamp)
			if precomputable then
				visited.target.accept(Current)
			end
			if precomputable then
				visited.arguments.accept(Current)
			end
		end

	visit_assignment (visited: ASSIGNMENT) is
		do
			visited.right_side.accept(Current)
		end

	visit_assignment_attempt (visited: ASSIGNMENT_ATTEMPT) is
		do
			precomputable := False -- *** TO DO
		end

	visit_ifthenelse (visited: IFTHENELSE) is
		local
			elseif_list: FAST_ARRAY[IFTHEN]; i: INTEGER
		do
			visited.expression.accept(Current)
			if precomputable and then visited.then_compound /= Void then
				visited.then_compound.accept(Current)
			end
			if precomputable and then visited.elseif_list /= Void then
				from
					elseif_list := visited.elseif_list
					i := elseif_list.upper
				until
					not precomputable or else i < elseif_list.lower
				loop
					visit_ifthen(elseif_list.item(i))
					i := i - 1
				end
			end
			if precomputable and then visited.else_compound /= Void then
				visited.else_compound.accept(Current)
			end
		end

	visit_ifthen (visited: IFTHEN) is
		do
			visited.expression.accept(Current)
			if precomputable and then visited.then_compound /= Void then
				visited.then_compound.accept(Current)
			end
		end

	visit_e_loop (visited: E_LOOP) is
		do
			if visited.initialize /= Void then
				visited.initialize.accept(Current)
			end
			if visited.loop_invariant /= Void then
				-- It may crash during precomputation:
				precomputable := False
			end
			if visited.loop_variant /= Void then
				-- It may crash during precomputation:
				precomputable := False
			end
			if precomputable then
				visited.until_expression.accept(Current)
			end
			if precomputable and then visited.loop_body /= Void then
				visited.loop_body.accept(Current)
			end
		end

	visit_inspect_statement (visited: INSPECT_STATEMENT) is
		local
			when_list: FAST_ARRAY[WHEN_CLAUSE]; i: INTEGER; when_clause: WHEN_CLAUSE
		do
			visited.expression.accept(Current)
			if precomputable and then visited.when_list /= Void then
				from
					when_list := visited.when_list
					i := when_list.upper
				until
					not precomputable or else i < when_list.lower
				loop
					when_clause := when_list.item(i)
					if when_clause.compound /= Void then
						when_clause.compound.accept(Current)
					end
					i := i - 1
				end
			end
			if precomputable and then visited.else_compound /= Void then
				visited.else_compound.accept(Current)
			end
		end

	visit_create_instruction (visited: CREATE_INSTRUCTION) is
		local
			call: PROC_CALL; created_type: TYPE; e_procedure: E_PROCEDURE
		do
			call := visited.call
			if call /= Void and then call.arguments /= Void then
				call.arguments.accept(Current)
			end
			if call /= Void and then  precomputable then
				created_type := visited.created_type(current_dynamic_type)
				e_procedure := visited.creation_procedure(created_type) -- Push:
				type_stack.add_last(created_type)
				anonymous_feature_stack.add_last(e_procedure)
				-- Visit the called one:
				e_procedure.accept(Current)
				-- Pop:
				anonymous_feature_stack.remove_last
				type_stack.remove_last
			end
		end

	visit_create_expression (visited: CREATE_EXPRESSION) is
		local
			call: PROC_CALL; created_type: TYPE; e_procedure: E_PROCEDURE
		do
			call := visited.call
			if call /= Void and then call.arguments /= Void then
				call.arguments.accept(Current)
			end
			if call /= Void and then precomputable then
				created_type := visited.created_type(current_dynamic_type)
				e_procedure := visited.creation_procedure(created_type) -- Push:
				type_stack.add_last(created_type)
				anonymous_feature_stack.add_last(e_procedure)
				-- Visit the called one:
				e_procedure.accept(Current)
				-- Pop:
				anonymous_feature_stack.remove_last
				type_stack.remove_last
			end
		end

	visit_implicit_cast (visited: IMPLICIT_CAST) is
		do
			-- Better to wait if it is actually simplified.
			precomputable := False
		end

	visit_integer_constant (visited: INTEGER_CONSTANT) is
		do
		end

	visit_character_constant (visited: CHARACTER_CONSTANT) is
		do
		end

	visit_real_constant (visited: REAL_CONSTANT) is
		do
		end

	visit_bit_constant (visited: BIT_CONSTANT) is
		do
		end

	visit_e_void (visited: E_VOID) is
		do
		end

	visit_null_pointer (visited: NULL_POINTER) is
		do
		end

	visit_manifest_string (visited: MANIFEST_STRING) is
		do
		end

	visit_e_true (visited: E_TRUE) is
		do
		end

	visit_e_false (visited: E_FALSE) is
		do
		end

	visit_written_current (visited: WRITTEN_CURRENT) is
		do
		end

	visit_implicit_current (visited: IMPLICIT_CURRENT) is
		do
		end

	visit_local_name1 (visited: LOCAL_NAME1) is
		do
		end

	visit_local_name2 (visited: LOCAL_NAME2) is
		do
		end

	visit_e_result (visited: E_RESULT) is
		do
		end

	visit_argument_name1 (visited: ARGUMENT_NAME1) is
		do
		end

	visit_argument_name2 (visited: ARGUMENT_NAME2) is
		do
		end

	visit_writable_attribute (visited: WRITABLE_ATTRIBUTE) is
		do
		end

	visit_feature_name (visited: FEATURE_NAME) is
		do
		end

	visit_void_call (visited: VOID_CALL) is
		do
			-- (Because the once routine may stay un-called.)
			precomputable := False
		end

	visit_void_proc_call (visited: VOID_PROC_CALL) is
		do
			-- (Because the once routine may stay un-called.)
			precomputable := False
		end

	visit_agent_expression (visited: AGENT_EXPRESSION) is
		do
			precomputable := False
		end

	visit_agent_instruction (visited: AGENT_INSTRUCTION) is
		do
			precomputable := False
		end

	visit_like_argument_wrapper (visited: LIKE_ARGUMENT_WRAPPER) is
		do
			precomputable := False
		end

	visit_writable (visited: WRITABLE) is
		do
		end

	visit_anonymous_feature (fs: FEATURE_STAMP; type: TYPE) is
		require
			fs /= Void
			type /= Void
		local
			af: ANONYMOUS_FEATURE
		do
			if fs /= fs.resolve_dynamic_binding_for(type) then
				-- No risk ;-)
				precomputable := False
			else
				af := fs.anonymous_feature(type)
				if anonymous_feature_stack.fast_has(af) then
					-- I do not trust recursive calls:
					precomputable := False
				else
					-- Push:
					type_stack.add_last(type)
					anonymous_feature_stack.add_last(af)
					-- Visit the called one:
					af.accept(Current)
					-- Pop:
					anonymous_feature_stack.remove_last
					type_stack.remove_last
				end
			end
		end

	visit_call_0 (visited: CALL_0) is
		do
			visit_call(visited.target, visited.feature_stamp)
			if precomputable then
				visited.target.accept(Current)
			end
		end

	visit_call_0_c (visited: CALL_0_C) is
		do
			visit_call_0(visited)
		end

	visit_call_1 (visited: CALL_1) is
		do
			visit_call(visited.target, visited.feature_stamp)
			if precomputable then
				visited.target.accept(Current)
			end
			if precomputable then
				visited.arg1.accept(Current)
			end
		end

	visit_call_n (visited: CALL_N) is
		do
			visit_call(visited.target, visited.feature_stamp)
			if precomputable then
				visited.target.accept(Current)
			end
			if precomputable then
				visited.arguments.accept(Current)
			end
		end

	visit_call_prefix_plus (visited: CALL_PREFIX_PLUS) is
		do
			visit_call_0(visited)
		end

	visit_call_prefix_minus (visited: CALL_PREFIX_MINUS) is
		do
			visit_call_0(visited)
		end

	visit_call_prefix_freeop (visited: CALL_PREFIX_FREEOP) is
		do
			visit_call_0(visited)
		end

	visit_call_prefix_not (visited: CALL_PREFIX_NOT) is
		do
			visit_call_0(visited)
		end

	visit_call_1_c (visited: CALL_1_C) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_plus (visited: CALL_INFIX_PLUS) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_minus (visited: CALL_INFIX_MINUS) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_freeop (visited: CALL_INFIX_FREEOP) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_gt (visited: CALL_INFIX_GT) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_ge (visited: CALL_INFIX_GE) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_lt (visited: CALL_INFIX_LT) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_le (visited: CALL_INFIX_LE) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_implies (visited: CALL_INFIX_IMPLIES) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_xor (visited: CALL_INFIX_XOR) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_or_else (visited: CALL_INFIX_OR_ELSE) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_and (visited: CALL_INFIX_AND) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_and_then (visited: CALL_INFIX_AND_THEN) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_int_div (visited: CALL_INFIX_INT_DIV) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_int_rem (visited: CALL_INFIX_INT_REM) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_div (visited: CALL_INFIX_DIV) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_times (visited: CALL_INFIX_TIMES) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_power (visited: CALL_INFIX_POWER) is
		do
			visit_call_1(visited)
		end

	visit_call_infix_or (visited: CALL_INFIX_OR) is
		do
			visit_call_1(visited)
		end

	visit_built_in_eq_neq (visited: BUILT_IN_EQ_NEQ) is
		do
			visited.left_side.accept(Current)
			visited.right_side.accept(Current)
		end

	visit_open_operand (visited: OPEN_OPERAND) is
		do
			precomputable := False
		end

	visit_expression_with_comment (visited: EXPRESSION_WITH_COMMENT) is
		do
			precomputable := False
		end

	visit_e_old (visited: E_OLD) is
		do
			precomputable := False
		end

	visit_fake_target (visited: FAKE_TARGET) is
		do
			precomputable := False
		end

	visit_fake_argument (visited: FAKE_ARGUMENT) is
		do
			precomputable := False
		end

	visit_agent_creation (visited: AGENT_CREATION) is
		do
			precomputable := False
		end

	visit_address_of (visited: ADDRESS_OF) is
		do
			precomputable := False
		end

	visit_non_void_no_dispatch (visited: NON_VOID_NO_DISPATCH) is
		do
			-- (This can be another precomputed once :-)
		end

feature {ONCE_ROUTINE_POOL}
	visit (cdt: like current_dynamic_type; once_body: INSTRUCTION): BOOLEAN is
		require
			cdt /= Void
			once_body /= Void
		do
			precomputable := True
			type_stack.clear_count
			type_stack.add_last(cdt)
			anonymous_feature_stack.clear_count
			once_body.accept(Current)
			Result := precomputable
		end

feature {}
	type_stack: FAST_ARRAY[TYPE] is
			-- To push and pop the `current_dynamic_type'.
		once
			create Result.with_capacity(32)
		end

	anonymous_feature_stack: FAST_ARRAY[ANONYMOUS_FEATURE] is
			-- To avoid infinite recursive calls.
		once
			create Result.with_capacity(32)
		end

	current_dynamic_type: TYPE is
			-- Of the currenly visited feature.
		do
			Result := type_stack.last
		end

	precomputable: BOOLEAN
			-- Memorize the result of the `visit'.

	make is
		do
		end

end -- class PRECOMPUTABLE_ROUTINE_DETECTOR
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
