-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
expanded class TMP_FEATURE
	--
	-- Temporary object used during syntax analysis.
	-- At the end, the good effective ANONYMOUS_FEATURE is choose.
	--

inherit
	ANY
insert
	GLOBALS

feature {EIFFEL_PARSER}
	arguments: FORMAL_ARG_LIST

	type: TYPE_MARK

	header_comment: COMMENT

	obsolete_mark: MANIFEST_STRING

	require_assertion: E_REQUIRE

	local_vars: LOCAL_VAR_LIST

	routine_body: INSTRUCTION

	initialize is
		do
			names.clear_count
			arguments := Void
			type := Void
			header_comment := Void
			obsolete_mark := Void
			require_assertion := Void
			local_vars := Void
			routine_body := Void
		end

	add_synonym (a_name: FEATURE_NAME) is
		require
			a_name /= Void
		do
			if a_name.to_string = as_void then
				error_handler.add_position(a_name.start_position)
				error_handler.append("Feature `Void' cannot be redefined (builtin).")
				error_handler.print_as_fatal_error
			end
			names.add_last(a_name)
		end

	set_arguments (args: like arguments) is
		require
			args /= Void
		do
			arguments := args
		end

	set_type (t: like type) is
		require
			t /= Void
		do
			type := t
		ensure
			type = t
		end

	set_header_comment (hc: like header_comment) is
		do
			header_comment := hc
		end

	set_obsolete_mark (om: like obsolete_mark) is
		do
			obsolete_mark := om
		end

	set_local_vars (lv: like local_vars) is
		do
			local_vars := lv
		end

	set_require (sp: POSITION; else_flag: BOOLEAN; hc: COMMENT; al: FAST_ARRAY[ASSERTION]) is
		do
			if hc /= Void or else al /= Void then
				create require_assertion.make(sp, else_flag, hc, al)
			end
		end

	set_routine_body (rb: like routine_body) is
		do
			routine_body := rb
		end

	as_writable_attribute: FEATURE_TEXT is
		do
			if type = Void then
				error_handler.add_position(names.first.start_position)
				error_handler.append("Bad feature definition.")
				error_handler.print_as_fatal_error
			elseif arguments /= Void then
				error_handler.add_position(names.first.start_position)
				error_handler.append("Attribute must not have formal arguments (or check if the %"is%" keyword is not missing).")
				error_handler.print_as_fatal_error
			end
			create Result.writable_attribute(n, type)
		end

	as_boolean_constant (value: BOOLEAN_CONSTANT): FEATURE_TEXT is
		do
			constant_result_type_check
			if type.is_boolean then
				create Result.boolean_constant(n, type, value)
			else
				error_handler.add_position(type.start_position)
				error_handler.append("The type of this constant feature should be BOOLEAN.")
				error_handler.print_as_fatal_error
			end
		end

	as_bit_constant (value: BIT_CONSTANT): FEATURE_TEXT is
		do
			constant_result_type_check
			if not type.is_bit then
				error_handler.add_position(type.start_position)
				error_handler.append("The type of this constant feature should be BIT.")
				error_handler.print_as_fatal_error
			end
			create Result.bit_constant(n, type, value)
		end

	as_character_constant (value: CHARACTER_CONSTANT): FEATURE_TEXT is
		do
			constant_result_type_check
			if not type.is_character then
				error_handler.add_position(type.start_position)
				error_handler.append("The type of this constant feature should be CHARACTER.")
				error_handler.print_as_fatal_error
			end
			create Result.character_constant(n, type, value)
		end

	as_integer_constant (value: INTEGER_CONSTANT): FEATURE_TEXT is
		local
			declaration_type: TYPE_INTEGER; value_memory: INTEGER_64; fatal_error_flag: BOOLEAN
		do
			constant_result_type_check
			if type.is_integer then
				value_memory := value.value_memory
				declaration_type ?= type
				inspect
					declaration_type.c_sizeof
				when 1 then
					fatal_error_flag := not value_memory.fit_integer_8
				when 2 then
					fatal_error_flag := not value_memory.fit_integer_16
				when 4 then
					fatal_error_flag := not value_memory.fit_integer_32
				when 8 then
				end
				if fatal_error_flag then
					error_handler.add_position(value.start_position)
					error_handler.append("Bad constant attribute definition (manifest constant value out of range).")
					error_handler.print_as_fatal_error
				else
					create Result.integer_constant(n, type, value)
				end
			elseif type.is_real then
				create Result.real_constant(n, type, value.to_real_constant)
			elseif type.is_double then
				create Result.double_constant(n, type, value.to_real_constant)
			else
				error_handler.add_position(type.start_position)
				error_handler.append("The type of this constant feature should be INTEGER REAL or DOUBLE.")
				error_handler.print_as_fatal_error
			end
		end

	as_real_constant (value: REAL_CONSTANT): FEATURE_TEXT is
		do
			constant_result_type_check
			if type.is_real then
				create Result.real_constant(n, type, value)
			elseif type.is_double then
				value.set_result_type(type)
				create Result.double_constant(n, type, value)
			else
				error_handler.add_position(type.start_position)
				error_handler.append("The type of this constant feature should be REAL or DOUBLE.")
				error_handler.print_as_fatal_error
			end
		end

	as_string_constant (value: MANIFEST_STRING): FEATURE_TEXT is
		do
			constant_result_type_check
			if not type.is_string then
				error_handler.add_position(type.start_position)
				error_handler.append("The type of this constant feature should be STRING.")
				error_handler.print_as_fatal_error
			end
			create Result.string_constant(n, type, value)
		end

	as_deferred_routine: FEATURE_TEXT is
		do
			if type = Void then
				create Result.deferred_procedure(n, arguments, obsolete_mark, header_comment, require_assertion)
			else
				create Result.deferred_function(n, arguments, type, obsolete_mark, header_comment, require_assertion)
			end
		end

	as_external_routine (native: NATIVE; alias_tag: MANIFEST_STRING): FEATURE_TEXT is
		do
			if type = Void then
				create Result.external_procedure(n, arguments, obsolete_mark, header_comment, require_assertion, native, alias_tag)
			else
				create Result.external_function(n, arguments, type, obsolete_mark, header_comment, require_assertion, native, alias_tag)
			end
		end

	as_once_routine: FEATURE_TEXT is
		do
			if type = Void then
				create Result.once_procedure(n, arguments, obsolete_mark, header_comment, require_assertion, local_vars, routine_body)
			else
				create Result.once_function(n, arguments, type, obsolete_mark, header_comment, require_assertion, local_vars, routine_body)
			end
		end

	as_procedure_or_function: FEATURE_TEXT is
		do
			if type = Void then
				create Result.e_procedure(n, arguments, obsolete_mark, header_comment, require_assertion, local_vars, routine_body)
			else
				create Result.e_function(n, arguments, type, obsolete_mark, header_comment, require_assertion, local_vars, routine_body)
			end
		end

	as_unique_constant: FEATURE_TEXT is
		do
			constant_result_type_check
			if not type.is_integer then
				error_handler.add_position(names.first.start_position)
				error_handler.append("Unique feature must have INTEGER type.")
				error_handler.print_as_fatal_error
			end
			create Result.unique_constant(n, type)
		end

feature {}
	names: FAST_ARRAY[FEATURE_NAME] is
		once
			create Result.make(8)
		end

	constant_result_type_check is
		do
			if type = Void then
				error_handler.add_position(names.first.start_position)
				error_handler.append("Bad constant declaration (no result type).")
				error_handler.print_as_fatal_error
			elseif not type.is_static then
				error_handler.add_position(type.start_position)
				error_handler.append("Must not use such a non-static type mark for constant.")
				error_handler.print_as_fatal_error
			end
		end

	n: FEATURE_NAME_LIST is
		require
			not names.is_empty
		do
			create Result.make_n(names)
		end

end -- class TMP_FEATURE
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
