-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class STRING_ALIASER
	--
	-- Singleton object used to share constant immutable strings (simple STRING
	-- as well as HASHED_STRING). The goal is to optimize immutable strings
	-- comparisons as well as hashed dictionary access (see HASHED_STRING).
	-- Thus, any immutable name must be registered here to get the
	-- corresponding unique reference. This singleton is shared via the
	-- `string_aliaser' once function from GLOBALS.
	--

inherit
	VISITABLE
insert
	GLOBALS

creation {ANY}
	make

feature {ANY}
	registered_one (model: STRING): BOOLEAN is
			-- Is the `model' _the_ already registered one? (i.e. True only if
			-- `model' has been returned by some previous `item' call).
		require
			model /= Void
		local
			hs, buffer: like hashed_string_buffer; hash_code: INTEGER
		do
			buffer := hashed_string_buffer
			buffer.set(model)
			hash_code := buffer.hash_code
			hs := memory.reference_at(buffer)
			if hs /= Void then
				check
					hs.to_string.is_equal(model)
				end
				Result := hs.to_string = model
			end
		end

	hashed_string (model: STRING): HASHED_STRING is
		require
			model /= Void
		local
			buffer: like hashed_string_buffer; hash_code: INTEGER
		do
			buffer := hashed_string_buffer
			buffer.set(model)
			hash_code := buffer.hash_code
			Result := memory.reference_at(buffer)
			if Result = Void then
				create Result.make(model.twin, hash_code)
				check
					not memory.has(Result)
				end
				memory.add(Result)
			end
		ensure
			Result.to_string.is_equal(model)
		end

	string (model: STRING): STRING is
			-- First creation of the corresponding aliased STRING.
			-- Note: the constraining require assertion is here to try to avoid
			-- multiple aliasing of the very same STRING.
		require
			not registered_one(model)
		do
			Result := hashed_string(model).to_string
		ensure
			Result.is_equal(model)
		end

feature {ANY} -- Some common HASHED_STRING to be shared:
	tuple_name: HASHED_STRING is
		once
			Result := string_aliaser.hashed_string(as_tuple)
		end

feature {FEATURE_NAME}
	for_prefix (to_string: STRING): HASHED_STRING is
		do
			Result := key_for_prefix_infix(once "_px_", to_string)
		end

	for_infix (to_string: STRING): HASHED_STRING is
		do
			Result := key_for_prefix_infix(once "_ix_", to_string)
		end

feature {COMPILE_TO_C, COMPILE_TO_JVM}
	echo_information is
		do
			echo.put_string(once "Aliased STRINGs: ")
			echo.put_integer(memory.count)
			echo.put_string(fz_dot_new_line)
		end

feature {EIFFEL_PARSER}
	case_insensitive_notify is
		require
			eiffel_parser.case_insensitive
		do
			case_insensitive_for(as_boolean_bits)
			case_insensitive_for(as_character_bits)
			case_insensitive_for(as_double_bits)
			case_insensitive_for(as_integer_bits)
			case_insensitive_for(as_minimum_character_code)
			case_insensitive_for(as_minimum_double)
			case_insensitive_for(as_minimum_real)
			case_insensitive_for(as_maximum_character_code)
			case_insensitive_for(as_maximum_double)
			case_insensitive_for(as_maximum_real)
			case_insensitive_for(as_pointer_bits)
			case_insensitive_for(as_real_bits)
		end

feature {ANY}
	accept (visitor: STRING_ALIASER_VISITOR) is
		do
			visitor.visit_string_aliaser(Current)
		end

feature {}
	make is
		do
			-- Hard-coded class names:
			memory_add(as_any)
			memory_add(as_array)
			memory_add(as_bit)
			memory_add(as_bit_n)
			memory_add(as_boolean)
			memory_add(as_boolean_ref)
			memory_add(as_character)
			memory_add(as_character_ref)
			memory_add(as_dictionary)
			memory_add(as_disposable)
			memory_add(as_double)
			memory_add(as_double_ref)
			memory_add(as_exceptions)
			memory_add(as_fixed_array)
			memory_add(as_function)
			memory_add(as_general)
			memory_add(as_hashed_dictionary)
			memory_add(as_hashed_set)
			memory_add(as_integer_general)
			memory_add(as_integer_8)
			memory_add(as_integer_16)
			memory_add(as_integer_32)
			memory_add(as_integer)
			memory_add(as_integer_64)
			memory_add(as_integer_ref)
			memory_add(as_native_array)
			memory_add(as_none)
			memory_add(as_platform)
			memory_add(as_pointer)
			memory_add(as_pointer_ref)
			memory_add(as_predicate)
			memory_add(as_procedure)
			memory_add(as_real_general)
			memory_add(as_real_32)
			memory_add(as_real)
			memory_add(as_real_64)
			memory_add(as_real_ref)
			memory_add(as_routine)
			memory_add(as_set)
			memory_add(as_string)
			memory_add(as_text_file_read)
			memory_add(as_text_file_write)
			memory_add(as_tuple)
			memory_add(as_type)
			memory_add(as_unicode_string)
			memory_add(as_weak_reference)
			-- Operator/Infix/Prefix list:
			memory_add(as_and)
			memory_add(as_and_then)
			memory_add(as_at)
			memory_add(as_backslash_backslash)
			memory_add(as_bit_and)
			memory_add(as_bit_clear)
			memory_add(as_bit_not)
			memory_add(as_bit_or)
			memory_add(as_bit_put)
			memory_add(as_bit_rotate)
			memory_add(as_bit_rotate_left)
			memory_add(as_bit_rotate_right)
			memory_add(as_bit_set)
			memory_add(as_bit_shift_right_unsigned)
			memory_add(as_bit_xor)
			memory_add(as_eq)
			memory_add(as_ge)
			memory_add(as_gt)
			memory_add(as_implies)
			memory_add(as_le)
			memory_add(as_lt)
			memory_add(as_minus)
			memory_add(as_muls)
			memory_add(as_neq)
			memory_add(as_not)
			memory_add(as_or)
			memory_add(as_or_else)
			memory_add(as_plus)
			memory_add(as_pow)
			memory_add(as_sharp_backslash_backslash)
			memory_add(as_sharp_minus)
			memory_add(as_sharp_muls)
			memory_add(as_sharp_plus)
			memory_add(as_sharp_slash_slash)
			memory_add(as_shift_left)
			memory_add(as_shift_right)
			memory_add(as_slash)
			memory_add(as_slash_slash)
			memory_add(as_xor)
			-- Hard coded feature names:
			memory_add(as_a1)
			memory_add(as_add_last)
			memory_add(as_agent)
			memory_add(as_as_16_ne)
			memory_add(as_atan2)
			memory_add(as_blank)
			memory_add(as_boolean_bits)
			memory_add(as_call)
			memory_add(as_calloc)
			memory_add(as_capacity)
			memory_add(as_character_bits)
			memory_add(as_clear_all)
			memory_add(as_count)
			memory_add(as_code)
			memory_add(as_collecting)
			memory_add(as_collection_off)
			memory_add(as_collection_on)
			memory_add(as_collector_counter)
			memory_add(as_copy)
			memory_add(as_c_inline_c)
			memory_add(as_c_inline_h)
			memory_add(as_default_create)
			memory_add(as_default_rescue)
			memory_add(as_deep_memcmp)
			memory_add(as_deep_twin)
			memory_add(as_deep_twin_from)
			memory_add(as_dispose)
			memory_add(as_double_bits)
			memory_add(as_double_floor)
			memory_add(as_die_with_code)
			memory_add(as_element_sizeof)
			memory_add(as_exception)
			memory_add(as_fifth)
			memory_add(as_first)
			memory_add(as_floor)
			memory_add(as_flush)
			memory_add(as_fourth)
			memory_add(as_from_pointer)
			memory_add(as_full_collect)
			memory_add(as_generating_type)
			memory_add(as_generator)
			memory_add(as_io)
			memory_add(as_integer_bits)
			memory_add(as_is_basic_expanded_type)
			memory_add(as_is_deep_equal)
			memory_add(as_is_expanded_type)
			memory_add(as_is_equal)
			memory_add(as_is_not_null)
			memory_add(as_item)
			memory_add(as_last)
			memory_add(as_lower)
			memory_add(as_low_8)
			memory_add(as_low_16)
			memory_add(as_low_32)
			memory_add(as_make)
			memory_add(as_minimum_character_code)
			memory_add(as_minimum_double)
			memory_add(as_minimum_real)
			memory_add(as_manifest_initialize)
			memory_add(as_maximum_character_code)
			memory_add(as_maximum_double)
			memory_add(as_maximum_real)
			memory_add(as_object_size)
			memory_add(as_pointer_bits)
			memory_add(as_pointer_size)
			memory_add(as_pow_postfix)
			memory_add(as_print)
			memory_add(as_print_on)
			memory_add(as_print_run_time_stack)
			memory_add(as_print_all_run_time_stacks)
			memory_add(as_put)
			memory_add(as_put_0)
			memory_add(as_put_1)
			memory_add(as_put_16_be)
			memory_add(as_put_16_le)
			memory_add(as_put_16_ne)
			memory_add(as_raise_exception)
			memory_add(as_real_bits)
			memory_add(as_realloc)
			memory_add(as_second)
			memory_add(as_set_item)
			memory_add(as_same_dynamic_type)
			memory_add(as_se_argc)
			memory_add(as_se_argv)
			memory_add(as_se_atexit)
			memory_add(as_sedb_breakpoint)
			memory_add(as_signal_number)
			memory_add(as_standard_copy)
			memory_add(as_standard_is_equal)
			memory_add(as_standard_twin)
			memory_add(as_std_error)
			memory_add(as_std_input)
			memory_add(as_std_output)
			memory_add(as_stderr)
			memory_add(as_stdin)
			memory_add(as_stdout)
			memory_add(as_storage)
			memory_add(as_third)
			memory_add(as_to_bit)
			memory_add(as_to_character)
			memory_add(as_to_double)
			memory_add(as_to_integer_8)
			memory_add(as_to_integer_16)
			memory_add(as_to_integer)
			memory_add(as_to_integer_32)
			memory_add(as_to_pointer)
			memory_add(as_to_real)
			memory_add(as_trace_switch)
			memory_add(as_truncated_to_integer)
			memory_add(as_twin)
			memory_add(as_upper)
			memory_add(as_with_capacity)
			-- Other names:
			memory_add(as_arguments)
			memory_add(as_current)
			memory_add(as_native_array_character)
			memory_add(as_last_result)
			memory_add(as_like_current)
			memory_add(as_method)
			memory_add(as_open_arguments)
			memory_add(as_open_argument_indices)
			memory_add(as_open_argument_index)
			memory_add(as_open_argument_count)
			memory_add(as_precursor)
			memory_add(as_result)
			memory_add(as_target)
			memory_add(as_void)
		end

	key_for_prefix_infix (tag_prefix_or_infix, to_string: STRING): HASHED_STRING is
		local
			i: INTEGER; c: CHARACTER; buffer: STRING
		do
			buffer := once ".....         local unique buffer          ....."
			buffer.copy(tag_prefix_or_infix)
			from
				i := 1
			until
				i > to_string.count
			loop
				c := to_string.item(i)
				if c.is_letter then
					buffer.extend(c)
				else
					c.code.append_in(buffer)
				end
				i := i + 1
			end
			Result := hashed_string(buffer)
		end

	memory_add (str: STRING) is
		require
			not registered_one(str)
		local
			hs: HASHED_STRING
		do
			create hs.make(str, str.hash_code)
			check
				not memory.has(hs)
			end
			memory.add(hs)
		end

	case_insensitive_for (str: STRING) is
		require
			registered_one(str)
		local
			hs, buffer: like hashed_string_buffer
		do
			buffer := hashed_string_buffer
			buffer.set(str)
			hs := memory.reference_at(buffer)
			check
				hs /= Void
				str = hs.to_string
			end
			memory.remove(hs)
			str.to_lower
			hs.set(str)
			memory.add(hs)
		end

	hashed_string_buffer: HASHED_STRING is
		once
			create Result.make(as_any, as_any.hash_code)
		end

	memory: SET[HASHED_STRING] is
		once
			create {HASHED_SET[HASHED_STRING]} Result.with_capacity(16384)
		end

	singleton_memory: STRING_ALIASER is
		once
			Result := Current
		end

invariant
	is_real_singleton: Current = singleton_memory

end -- class STRING_ALIASER
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
