-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class EXPANDED_TYPE_MARK
	--
	-- Handling of the "expanded FOO" type mark.
	--

inherit
	TYPE_MARK

creation {ANY}
	make

feature {ANY}
	is_expanded: BOOLEAN is True

	is_reference: BOOLEAN is False

	jvm_method_flags: INTEGER is 17

	start_position: POSITION
			-- Of first letter of keyword `expanded'.

	mapped: TYPE_MARK
			-- Corresponding mapped type (usualy the reference type).

	written_name: HASHED_STRING

	is_static: BOOLEAN is
		do
			if mapped.is_static then
				Result := not mapped.is_expanded
			end
		end

	is_separate: BOOLEAN is
		do
			Result := mapped.is_separate
		end

	is_generic: BOOLEAN is
		do
			Result := mapped.is_generic
		end

	generic_list: ARRAY[TYPE_MARK] is
		do
			Result := mapped.generic_list
		end

	specialize_in (new_type: TYPE): like Current is
		do
			Result := current_or_twin_init(mapped.specialize_in(new_type))
		end

	specialize_thru (parent_type: TYPE; inherit_item: INHERIT_TEXT_ITEM; new_type: TYPE): like Current is
		do
			Result := current_or_twin_init(mapped.specialize_thru(parent_type, inherit_item, new_type))
		end

	has_been_specialized: BOOLEAN is
		do
			Result := mapped.has_been_specialized
		end

	can_be_assigned_to (other: TYPE_MARK; t, o: TYPE; tp, op: POSITION): BOOLEAN is
		do
			Result := mapped.type.is_a(o, tp, op) --*** perf: type lookup
			if not Result then
				error_handler.conformance_error(t, o, tp, op)
			end
		end

	declaration_type: TYPE_MARK is
		local
			mdt: TYPE_MARK
		do
			if declaration_type_memory = Void then
				if mapped.is_static then
					if mapped.is_expanded then
						declaration_type_memory := mapped
					else
						declaration_type_memory := Current
					end
				else
					mdt := mapped.declaration_type
					if mdt.is_expanded then
						declaration_type_memory := mdt
					else
						create {EXPANDED_TYPE_MARK} declaration_type_memory.make(start_position, mdt)
					end
				end
			end
			Result := declaration_type_memory
		end

	constrained_type: TYPE_MARK is
		local
			mct: TYPE_MARK
		do
			--*** Can there actually be outside expanded types in a 
			--*** formal generic list ?
			--*** Could use a `constrained_type_memory' ? <FM-3/8/2004>
			if mapped.is_static then
				if mapped.is_expanded then
					Result := mapped
				else
					Result := Current
				end
			else
				mct := mapped.constrained_type
				if mct.is_expanded then
					Result := mct
				else
					create {EXPANDED_TYPE_MARK} Result.make(start_position, mct)
				end
			end
		end

	actual_reference (destination: TYPE_MARK): TYPE_MARK is
		local
			sp: POSITION
		do
			if mapped.is_expanded then
				sp := start_position
				if sp.is_unknown then
					if destination /= Void then
						sp := destination.start_position
					end
				end
				create {TYPE_REFERENCE} Result.make(sp, mapped)
			else
				Result := mapped
			end
		end

	actual_separate (destination: TYPE_MARK): TYPE_MARK is
		do
			check
				False
			end
		end

	local_from_separate: TYPE_MARK is
		do
			check
				False
			end
		end

	is_user_expanded: BOOLEAN is
		do
			Result := not mapped.is_kernel_expanded
		end

	is_dummy_expanded: BOOLEAN is
		do
			if is_user_expanded then
				Result := live_type.writable_attributes = Void
			end
		end

	id: INTEGER is
		do
			Result := live_type.id
		end

	smallest_ancestor (other: TYPE_MARK): TYPE_MARK is
		do
			if written_name = other.written_name then
				Result := Current
			else
				Result := mapped.smallest_ancestor(other)
			end
		end

	is_a (other: TYPE_MARK): BOOLEAN is
		do
			Result := mapped.is_a(other)
			if not Result then
				error_handler.type_error(Current, other)
			end
		end

	c_header_pass1 is
		do
			standard_c_typedef
		end

	c_header_pass2 is
		do
		end

	c_header_pass3 is
		do
			if need_c_struct then
				standard_c_struct
			end
			standard_c_object_model
		end

	c_header_pass4 is
		do
			standard_c_print_function
		end

	c_type_for_argument_in (str: STRING) is
		do
			if mapped.is_expanded then
				mapped.c_type_for_argument_in(str)
			elseif is_dummy_expanded then
				str.append(fz_int)
			else
				str.extend('T')
				id.append_in(str)
			end
		end

	c_type_for_target_in (str: STRING) is
		do
			if is_dummy_expanded then
				str.append(fz_int)
			else
				str.extend('T')
				id.append_in(str)
				str.extend('*')
			end
		end

	c_type_for_result_in (str: STRING) is
		do
			if is_dummy_expanded then
				str.append(fz_int)
			else
				str.extend('T')
				id.append_in(str)
			end
		end

	need_c_struct: BOOLEAN is
		do
			if mapped.is_expanded then
				Result := mapped.need_c_struct
			else
				Result := live_type.writable_attributes /= Void
			end
		end

	c_sizeof: INTEGER is
		do
			Result := live_type.c_sizeof
		end

	class_text_name: CLASS_NAME is
		do
			Result := mapped.class_text_name
		end

	c_initialize is
		do
			if is_kernel_expanded then
				mapped.c_initialize
			else
				c_initialize_expanded
			end
		end

	c_initialize_in (str: STRING) is
		do
			if is_kernel_expanded then
				mapped.c_initialize_in(str)
			else
				if need_c_struct then
					live_type.c_object_model_in(str)
				else
					str.extend('0')
				end
			end
		end

	jvm_descriptor_in (str: STRING) is
		do
			live_type.jvm_type_descriptor_in(str)
		end

	jvm_target_descriptor_in (str: STRING) is
		do
		end

	jvm_return_code is
		do
			code_attribute.opcode_areturn
		end

	jvm_push_local (offset: INTEGER) is
		do
			code_attribute.opcode_aload(offset)
		end

	jvm_check_class_invariant is
		do
			standard_jvm_check_class_invariant
		end

	jvm_push_default: INTEGER is
		do
			if is_kernel_expanded then
				Result := jvm_push_default
			else
				live_type.jvm_expanded_push_default
				Result := 1
			end
		end

	jvm_write_local_creation (offset: INTEGER) is
		do
			if is_kernel_expanded then
				jvm_write_local_creation(offset)
			else
				code_attribute.opcode_astore(offset)
			end
		end

	jvm_write_local (offset: INTEGER) is
		do
			if is_kernel_expanded then
				jvm_write_local(offset)
			else
				jvm_write_local_expanded(offset)
			end
		end

	jvm_write_local_expanded (offset: INTEGER) is
		local
			rc: LIVE_TYPE; wa: ARRAY[RUN_FEATURE_2]
		do
			rc := live_type
			wa := rc.writable_attributes
			code_attribute.opcode_aload(offset)
			code_attribute.opcode_swap
			jvm.fields_by_fields_expanded_copy(wa)
			code_attribute.opcode_pop2
		end

	jvm_xnewarray is
		local
			idx: INTEGER
		do
			if is_kernel_expanded then
				jvm_xnewarray
			else
				check
					is_user_expanded
				end
				idx := live_type.jvm_constant_pool_index
				code_attribute.opcode_anewarray(idx)
			end
		end

	jvm_xastore is
		do
			code_attribute.opcode_aastore
		end

	jvm_xaload is
		do
			code_attribute.opcode_aaload
		end

	jvm_if_x_eq: INTEGER is
		do
			Result := code_attribute.opcode_if_acmpeq
		end

	jvm_if_x_ne: INTEGER is
		do
			Result := code_attribute.opcode_if_acmpne
		end

	jvm_to_reference is
		local
			wa: ARRAY[RUN_FEATURE_2]
		do
			if mapped.is_expanded then
				live_type.jvm_to_reference
			else
				mapped.live_type.jvm_basic_new
				wa := live_type.writable_attributes
				code_attribute.opcode_swap
				jvm.fields_by_fields_copy(wa)
			end
		end

	jvm_expanded_from_reference (other: TYPE_MARK): INTEGER is
		do
			check
				False
			end
		end

	jvm_convert_to (destination: TYPE_MARK): INTEGER is
		do
			if destination.is_reference then
				jvm_to_reference
			end
			Result := 1
		end

	jvm_standard_is_equal is
		local
			rc: LIVE_TYPE; wa: ARRAY[RUN_FEATURE_2]
		do
			rc := live_type
			wa := rc.writable_attributes
			jvm.std_is_equal(rc, wa)
		end

	signature_resolve_in (new_type: TYPE): TYPE is
		do
			Result := resolve_in(new_type)
		end

	to_static (new_type: TYPE): TYPE_MARK is
		local
			m: like mapped
		do
			m := mapped.to_static(new_type)
			if m = mapped then
				if mapped.is_expanded then
					Result := mapped
				else
					Result := Current
				end
			elseif m.is_expanded then
				Result := m
			else
				create {EXPANDED_TYPE_MARK} Result.make(start_position, m)
			end
		end

	accept (visitor: EXPANDED_TYPE_MARK_VISITOR) is
		do
			visitor.visit_expanded_type_mark(Current)
		end

	need_gc_mark_function: BOOLEAN is
		do
			Result := live_type.gc_mark_to_follow
		end

feature {EXPANDED_TYPE_MARK}
	init (m: like mapped) is
		require
			m /= Void
		do
			mapped := m
		ensure
			mapped = m
		end

feature {}
	declaration_type_memory: TYPE_MARK

feature {LIVE_TYPE, TYPE_MARK}
	just_before_gc_mark_in (str: STRING) is
		do
		end

	gc_info_in (str: STRING) is
		do
		end

	gc_define1 is
		do
		end

	gc_define2 is
		do
			standard_gc_define2_for_expanded
		end

feature {TYPE_MARK}
	short_ (shorted_type: TYPE) is
		do
			short_printer.hook_or(fz_expanded, "expanded ")
			mapped.short_(shorted_type)
		end

feature {}
	make (sp: like start_position; m: like mapped) is
		require
			not sp.is_unknown
			m /= Void
		local
			mapped_wm, buffer: STRING
		do
			start_position := sp
			mapped := m -- Compute `written_name':
			mapped_wm := mapped.written_mark
			buffer := once "...This is a local once buffer..."
			buffer.copy(fz_expanded)
			buffer.extend(' ')
			buffer.append(mapped_wm)
			written_name := string_aliaser.hashed_string(buffer)
		ensure
			start_position = sp
			mapped = m
		end

	current_or_twin_init (m: like mapped): like Current is
		require
			m /= Void
		do
			if m = mapped then
				Result := Current
			else
				Result := twin
				Result.init(m)
			end
		ensure
			Result.mapped = m
		end

end -- class EXPANDED_TYPE_MARK
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
